import json
import logging

from unittest import mock

from django.conf import settings
from django.contrib.auth.models import Group, Permission
from django.core import mail
from django.core.mail import EmailMultiAlternatives
from django.test import TestCase, override_settings
from django.urls import reverse
from freezegun import freeze_time

from wagtail.admin.admin_url_finder import AdminURLFinder
from wagtail.models import (
    GroupApprovalTask, Page, Task, TaskState, UserProfile, Workflow, WorkflowPage, WorkflowState,
    WorkflowTask)
from wagtail.signals import page_published
from wagtail.test.testapp.models import SimplePage, SimpleTask
from wagtail.test.utils import WagtailTestUtils


def delete_existing_workflows():
    WorkflowPage.objects.all().delete()
    Workflow.objects.all().delete()
    Task.objects.all().delete()
    WorkflowTask.objects.all().delete()


class TestWorkflowMenus(TestCase, WagtailTestUtils):
    def setUp(self):
        self.login()

        self.editor = self.create_user(
            username='editor',
            email='editor@email.com',
            password='password',
        )
        editors = Group.objects.get(name='Editors')
        editors.user_set.add(self.editor)

    def test_workflow_settings_and_reports_menus_are_shown_to_admin(self):
        response = self.client.get('/admin/')
        self.assertContains(response, 'href="/admin/workflows/list/"')
        self.assertContains(response, 'href="/admin/workflows/tasks/index/"')
        self.assertContains(response, 'href="/admin/reports/workflow/"')
        self.assertContains(response, 'href="/admin/reports/workflow_tasks/"')

    def test_workflow_settings_menus_are_not_shown_to_editor(self):
        self.login(user=self.editor)
        response = self.client.get('/admin/')
        self.assertNotContains(response, 'href="/admin/workflows/list/"')
        self.assertNotContains(response, 'href="/admin/workflows/tasks/index/"')
        self.assertContains(response, 'href="/admin/reports/workflow/"')
        self.assertContains(response, 'href="/admin/reports/workflow_tasks/"')

    @override_settings(WAGTAIL_WORKFLOW_ENABLED=False)
    def test_workflow_menus_are_hidden_when_workflows_are_disabled(self):
        response = self.client.get('/admin/')
        self.assertNotContains(response, 'href="/admin/workflows/list/"')
        self.assertNotContains(response, 'href="/admin/workflows/tasks/index/"')
        self.assertNotContains(response, 'href="/admin/reports/workflow/"')
        self.assertNotContains(response, 'href="/admin/reports/workflow_tasks/"')


class TestWorkflowsIndexView(TestCase, WagtailTestUtils):

    def setUp(self):
        delete_existing_workflows()
        self.login()

        self.editor = self.create_user(
            username='editor',
            email='editor@email.com',
            password='password',
        )
        editors = Group.objects.get(name='Editors')
        editors.user_set.add(self.editor)

        self.moderator = self.create_user(
            username='moderator',
            email='moderator@email.com',
            password='password',
        )
        moderators = Group.objects.get(name='Moderators')
        moderators.user_set.add(self.moderator)
        moderators.permissions.add(Permission.objects.get(codename="add_workflow"))

    def get(self, params={}):
        return self.client.get(reverse('wagtailadmin_workflows:index'), params)

    def test_simple(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtailadmin/workflows/index.html')

        # Initially there should be no workflows listed
        self.assertContains(response, "There are no enabled workflows.")

        Workflow.objects.create(name="test_workflow", active=True)

        # Now the listing should contain our workflow
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtailadmin/workflows/index.html')
        self.assertNotContains(response, "There are no enabled workflows.")
        self.assertContains(response, "test_workflow")

    def test_deactivated(self):
        Workflow.objects.create(name="test_workflow", active=False)

        # The listing should contain our workflow, as well as marking it as disabled
        response = self.get(params={'show_disabled': 'true'})
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(response, "No workflows have been created.")
        self.assertContains(response, "test_workflow")
        self.assertContains(response, '<span class="status-tag">Disabled</span>', html=True)

        # If we set 'show_disabled' to 'False', the workflow should not be displayed
        response = self.get(params={})
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "There are no enabled workflows.")

    def test_permissions(self):
        self.login(user=self.editor)
        response = self.get()
        self.assertEqual(response.status_code, 302)
        full_context = {key: value for context in response.context for key, value in context.items()}
        self.assertEqual(full_context['message'], 'Sorry, you do not have permission to access this area.')

        self.login(user=self.moderator)
        response = self.get()
        self.assertEqual(response.status_code, 200)


class TestWorkflowsCreateView(TestCase, WagtailTestUtils):

    def setUp(self):
        delete_existing_workflows()
        self.login()
        self.task_1 = SimpleTask.objects.create(name="first_task")
        self.task_2 = SimpleTask.objects.create(name="second_task")

        self.editor = self.create_user(
            username='editor',
            email='editor@email.com',
            password='password',
        )
        editors = Group.objects.get(name='Editors')
        editors.user_set.add(self.editor)

        self.moderator = self.create_user(
            username='moderator',
            email='moderator@email.com',
            password='password',
        )
        moderators = Group.objects.get(name='Moderators')
        moderators.user_set.add(self.moderator)
        moderators.permissions.add(Permission.objects.get(codename="add_workflow"))

        self.root_page = Page.objects.get(depth=1)

    def get(self, params={}):
        return self.client.get(reverse('wagtailadmin_workflows:add'), params)

    def post(self, post_data={}):
        return self.client.post(reverse('wagtailadmin_workflows:add'), post_data)

    def test_get(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtailadmin/workflows/create.html')

    def test_post(self):
        response = self.post({
            'name': ['test_workflow'],
            'active': ['on'],
            'workflow_tasks-TOTAL_FORMS': ['2'],
            'workflow_tasks-INITIAL_FORMS': ['0'],
            'workflow_tasks-MIN_NUM_FORMS': ['0'],
            'workflow_tasks-MAX_NUM_FORMS': ['1000'],
            'workflow_tasks-0-task': [str(self.task_1.id)],
            'workflow_tasks-0-id': [''],
            'workflow_tasks-0-ORDER': ['1'],
            'workflow_tasks-0-DELETE': [''],
            'workflow_tasks-1-task': [str(self.task_2.id)],
            'workflow_tasks-1-id': [''],
            'workflow_tasks-1-ORDER': ['2'],
            'workflow_tasks-1-DELETE': [''],
            'pages-TOTAL_FORMS': ['2'],
            'pages-INITIAL_FORMS': ['1'],
            'pages-MIN_NUM_FORMS': ['0'],
            'pages-MAX_NUM_FORMS': ['1000'],
            'pages-0-page': [str(self.root_page.id)],
            'pages-0-DELETE': [''],
            'pages-1-page': [''],
            'pages-1-DELETE': [''],
        })

        # Should redirect back to index
        self.assertRedirects(response, reverse('wagtailadmin_workflows:index'))

        # Check that the workflow was created
        workflows = Workflow.objects.filter(name="test_workflow", active=True)
        self.assertEqual(workflows.count(), 1)

        workflow = workflows.first()

        # Check that the tasks are associated with the workflow
        self.assertEqual([self.task_1.task_ptr, self.task_2.task_ptr], list(workflow.tasks))

        # Check that the tasks have sort_order set on WorkflowTask correctly
        self.assertEqual(WorkflowTask.objects.get(workflow=workflow, task=self.task_1.task_ptr).sort_order, 0)
        self.assertEqual(WorkflowTask.objects.get(workflow=workflow, task=self.task_2.task_ptr).sort_order, 1)

    def test_permissions(self):
        self.login(user=self.editor)
        response = self.get()
        self.assertEqual(response.status_code, 302)
        full_context = {key: value for context in response.context for key, value in context.items()}
        self.assertEqual(full_context['message'], 'Sorry, you do not have permission to access this area.')

        self.login(user=self.moderator)
        response = self.get()
        self.assertEqual(response.status_code, 200)

    def test_page_already_has_workflow_check(self):
        workflow = Workflow.objects.create(name="existing_workflow")
        WorkflowPage.objects.create(workflow=workflow, page=self.root_page)

        response = self.post({
            'name': ['test_workflow'],
            'active': ['on'],
            'workflow_tasks-TOTAL_FORMS': ['2'],
            'workflow_tasks-INITIAL_FORMS': ['0'],
            'workflow_tasks-MIN_NUM_FORMS': ['0'],
            'workflow_tasks-MAX_NUM_FORMS': ['1000'],
            'workflow_tasks-0-task': [str(self.task_1.id)],
            'workflow_tasks-0-id': [''],
            'workflow_tasks-0-ORDER': ['1'],
            'workflow_tasks-0-DELETE': [''],
            'workflow_tasks-1-task': [str(self.task_2.id)],
            'workflow_tasks-1-id': [''],
            'workflow_tasks-1-ORDER': ['2'],
            'workflow_tasks-1-DELETE': [''],
            'pages-TOTAL_FORMS': ['2'],
            'pages-INITIAL_FORMS': ['1'],
            'pages-MIN_NUM_FORMS': ['0'],
            'pages-MAX_NUM_FORMS': ['1000'],
            'pages-0-page': [str(self.root_page.id)],
            'pages-0-DELETE': [''],
            'pages-1-page': [''],
            'pages-1-DELETE': [''],
        })

        self.assertEqual(response.status_code, 200)
        self.assertFormsetError(response, 'pages_formset', 0, 'page', ["This page already has workflow 'existing_workflow' assigned."])


class TestWorkflowsEditView(TestCase, WagtailTestUtils):

    def setUp(self):
        delete_existing_workflows()
        self.login()
        self.workflow = Workflow.objects.create(name="workflow_to_edit")
        self.task_1 = SimpleTask.objects.create(name="first_task")
        self.task_2 = SimpleTask.objects.create(name="second_task")
        self.inactive_task = SimpleTask.objects.create(name="inactive_task", active=False)
        self.workflow_task = WorkflowTask.objects.create(workflow=self.workflow, task=self.task_1.task_ptr, sort_order=0)
        self.page = Page.objects.first()
        WorkflowPage.objects.create(workflow=self.workflow, page=self.page)

        self.editor = self.create_user(
            username='editor',
            email='editor@email.com',
            password='password',
        )
        editors = Group.objects.get(name='Editors')
        editors.user_set.add(self.editor)

        self.moderator = self.create_user(
            username='moderator',
            email='moderator@email.com',
            password='password',
        )
        moderators = Group.objects.get(name='Moderators')
        moderators.user_set.add(self.moderator)
        moderators.permissions.add(Permission.objects.get(codename="change_workflow"))

    def get(self, params={}):
        return self.client.get(reverse('wagtailadmin_workflows:edit', args=[self.workflow.id]), params)

    def post(self, post_data={}):
        return self.client.post(reverse('wagtailadmin_workflows:edit', args=[self.workflow.id]), post_data)

    def test_get(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtailadmin/workflows/edit.html')

        # Check that the list of pages has the page to which this workflow is assigned
        self.assertContains(response, self.page.title)

    def test_post(self):
        response = self.post({
            'name': [str(self.workflow.name)],
            'active': ['on'],
            'workflow_tasks-TOTAL_FORMS': ['2'],
            'workflow_tasks-INITIAL_FORMS': ['1'],
            'workflow_tasks-MIN_NUM_FORMS': ['0'],
            'workflow_tasks-MAX_NUM_FORMS': ['1000'],
            'workflow_tasks-0-task': [str(self.task_1.id)],
            'workflow_tasks-0-id': [str(self.workflow_task.id)],
            'workflow_tasks-0-ORDER': ['1'],
            'workflow_tasks-0-DELETE': [''],
            'workflow_tasks-1-task': [str(self.task_2.id)],
            'workflow_tasks-1-id': [''],
            'workflow_tasks-1-ORDER': ['2'],
            'workflow_tasks-1-DELETE': [''],
            'pages-TOTAL_FORMS': ['2'],
            'pages-INITIAL_FORMS': ['1'],
            'pages-MIN_NUM_FORMS': ['0'],
            'pages-MAX_NUM_FORMS': ['1000'],
            'pages-0-page': [str(self.page.id)],
            'pages-0-DELETE': [''],
            'pages-1-page': [''],
            'pages-1-DELETE': [''],
        })

        # Should redirect back to index
        self.assertRedirects(response, reverse('wagtailadmin_workflows:index'))

        # Check that the workflow was created
        workflows = Workflow.objects.filter(name="workflow_to_edit", active=True)
        self.assertEqual(workflows.count(), 1)

        workflow = workflows.first()

        # Check that the tasks are associated with the workflow
        self.assertEqual([self.task_1.task_ptr, self.task_2.task_ptr], list(workflow.tasks))

        # Check that the tasks have sort_order set on WorkflowTask correctly
        self.assertEqual(WorkflowTask.objects.get(workflow=workflow, task=self.task_1.task_ptr).sort_order, 0)
        self.assertEqual(WorkflowTask.objects.get(workflow=workflow, task=self.task_2.task_ptr).sort_order, 1)

    def test_permissions(self):
        self.login(user=self.editor)
        response = self.get()
        self.assertEqual(response.status_code, 302)
        full_context = {key: value for context in response.context for key, value in context.items()}
        self.assertEqual(full_context['message'], 'Sorry, you do not have permission to access this area.')

        self.login(user=self.moderator)
        response = self.get()
        self.assertEqual(response.status_code, 200)

    def test_admin_url_finder(self):
        editor_url_finder = AdminURLFinder(self.editor)
        self.assertEqual(editor_url_finder.get_edit_url(self.workflow), None)
        moderator_url_finder = AdminURLFinder(self.moderator)
        expected_url = '/admin/workflows/edit/%d/' % self.workflow.pk
        self.assertEqual(moderator_url_finder.get_edit_url(self.workflow), expected_url)

    def test_duplicate_page_check(self):
        response = self.post({
            'name': [str(self.workflow.name)],
            'active': ['on'],
            'workflow_tasks-TOTAL_FORMS': ['2'],
            'workflow_tasks-INITIAL_FORMS': ['1'],
            'workflow_tasks-MIN_NUM_FORMS': ['0'],
            'workflow_tasks-MAX_NUM_FORMS': ['1000'],
            'workflow_tasks-0-task': [str(self.task_1.id)],
            'workflow_tasks-0-id': [str(self.workflow_task.id)],
            'workflow_tasks-0-ORDER': ['1'],
            'workflow_tasks-0-DELETE': [''],
            'workflow_tasks-1-task': [str(self.task_2.id)],
            'workflow_tasks-1-id': [''],
            'workflow_tasks-1-ORDER': ['2'],
            'workflow_tasks-1-DELETE': [''],
            'pages-TOTAL_FORMS': ['2'],
            'pages-INITIAL_FORMS': ['1'],
            'pages-MIN_NUM_FORMS': ['0'],
            'pages-MAX_NUM_FORMS': ['1000'],
            'pages-0-page': [str(self.page.id)],
            'pages-0-DELETE': [''],
            'pages-1-page': [str(self.page.id)],
            'pages-1-DELETE': [''],
        })

        self.assertEqual(response.status_code, 200)
        self.assertFormsetError(response, 'pages_formset', None, None, ['You cannot assign this workflow to the same page multiple times.'])

    def test_pages_ignored_if_workflow_disabled(self):
        self.workflow.active = False
        self.workflow.save()
        self.workflow.workflow_pages.all().delete()

        response = self.post({
            'name': [str(self.workflow.name)],
            'active': ['on'],
            'workflow_tasks-TOTAL_FORMS': ['2'],
            'workflow_tasks-INITIAL_FORMS': ['1'],
            'workflow_tasks-MIN_NUM_FORMS': ['0'],
            'workflow_tasks-MAX_NUM_FORMS': ['1000'],
            'workflow_tasks-0-task': [str(self.task_1.id)],
            'workflow_tasks-0-id': [str(self.workflow_task.id)],
            'workflow_tasks-0-ORDER': ['1'],
            'workflow_tasks-0-DELETE': [''],
            'workflow_tasks-1-task': [str(self.task_2.id)],
            'workflow_tasks-1-id': [''],
            'workflow_tasks-1-ORDER': ['2'],
            'workflow_tasks-1-DELETE': [''],
            'pages-TOTAL_FORMS': ['2'],
            'pages-INITIAL_FORMS': ['1'],
            'pages-MIN_NUM_FORMS': ['0'],
            'pages-MAX_NUM_FORMS': ['1000'],
            'pages-0-page': [str(self.page.id)],
            'pages-0-DELETE': [''],
            'pages-1-page': [''],
            'pages-1-DELETE': [''],
        })

        # Should redirect back to index
        self.assertRedirects(response, reverse('wagtailadmin_workflows:index'))

        # Check that the pages weren't added to the workflow
        self.workflow.refresh_from_db()
        self.assertFalse(self.workflow.workflow_pages.exists())


class TestRemoveWorkflow(TestCase, WagtailTestUtils):
    fixtures = ['test.json']

    def setUp(self):
        delete_existing_workflows()
        self.login()
        self.workflow = Workflow.objects.create(name="workflow")
        self.page = Page.objects.first()
        WorkflowPage.objects.create(workflow=self.workflow, page=self.page)

        self.editor = self.create_user(
            username='editor',
            email='editor@email.com',
            password='password',
        )
        editors = Group.objects.get(name='Editors')
        editors.user_set.add(self.editor)

        self.moderator = self.create_user(
            username='moderator',
            email='moderator@email.com',
            password='password',
        )
        moderators = Group.objects.get(name='Moderators')
        moderators.user_set.add(self.moderator)
        moderators.permissions.add(Permission.objects.get(codename="change_workflow"))

    def post(self, post_data={}):
        return self.client.post(reverse('wagtailadmin_workflows:remove', args=[self.page.id, self.workflow.id]), post_data)

    def test_post(self):
        # Check that a WorkflowPage instance is removed correctly
        self.post()
        self.assertEqual(WorkflowPage.objects.filter(workflow=self.workflow, page=self.page).count(), 0)

    def test_no_permissions(self):
        self.login(user=self.editor)
        response = self.post()
        self.assertEqual(response.status_code, 302)

    def test_post_with_permission(self):
        self.login(user=self.moderator)
        response = self.post()
        self.assertEqual(response.status_code, 302)


class TestTaskIndexView(TestCase, WagtailTestUtils):

    def setUp(self):
        delete_existing_workflows()
        self.login()

        self.editor = self.create_user(
            username='editor',
            email='editor@email.com',
            password='password',
        )
        editors = Group.objects.get(name='Editors')
        editors.user_set.add(self.editor)

        self.moderator = self.create_user(
            username='moderator',
            email='moderator@email.com',
            password='password',
        )
        moderators = Group.objects.get(name='Moderators')
        moderators.user_set.add(self.moderator)
        moderators.permissions.add(Permission.objects.get(codename="change_task"))

    def get(self, params={}):
        return self.client.get(reverse('wagtailadmin_workflows:task_index'), params)

    def test_simple(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtailadmin/workflows/task_index.html')

        # Initially there should be no tasks listed
        self.assertContains(response, "There are no enabled tasks")

        SimpleTask.objects.create(name="test_task", active=True)

        # Now the listing should contain our task
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtailadmin/workflows/task_index.html')
        self.assertNotContains(response, "There are no enabled tasks")
        self.assertContains(response, "test_task")

    def test_deactivated(self):
        Task.objects.create(name="test_task", active=False)

        # The listing should contain our task, as well as marking it as disabled
        response = self.get(params={'show_disabled': 'true'})
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(response, "No tasks have been created.")
        self.assertContains(response, "test_task")
        self.assertContains(response, '<span class="status-tag">Disabled</span>', html=True)

        # The listing should not contain task if show_disabled query parameter is 'False'
        response = self.get(params={})
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "There are no enabled tasks")
        self.assertNotContains(response, "test_task")

    def test_permissions(self):
        self.login(user=self.editor)
        response = self.get()
        self.assertEqual(response.status_code, 302)
        full_context = {key: value for context in response.context for key, value in context.items()}
        self.assertEqual(full_context['message'], 'Sorry, you do not have permission to access this area.')

        self.login(user=self.moderator)
        response = self.get()
        self.assertEqual(response.status_code, 200)


class TestCreateTaskView(TestCase, WagtailTestUtils):

    def setUp(self):
        delete_existing_workflows()
        self.login()

        self.editor = self.create_user(
            username='editor',
            email='editor@email.com',
            password='password',
        )
        editors = Group.objects.get(name='Editors')
        editors.user_set.add(self.editor)

        self.moderator = self.create_user(
            username='moderator',
            email='moderator@email.com',
            password='password',
        )
        moderators = Group.objects.get(name='Moderators')
        moderators.user_set.add(self.moderator)
        moderators.permissions.add(Permission.objects.get(codename="add_task"))

    def get(self, url_kwargs=None, params={}):
        url_kwargs = url_kwargs or {}
        url_kwargs.setdefault('app_label', SimpleTask._meta.app_label)
        url_kwargs.setdefault('model_name', SimpleTask._meta.model_name)
        return self.client.get(reverse('wagtailadmin_workflows:add_task', kwargs=url_kwargs), params)

    def post(self, post_data={}):
        return self.client.post(reverse('wagtailadmin_workflows:add_task', kwargs={'app_label': SimpleTask._meta.app_label, 'model_name': SimpleTask._meta.model_name}), post_data)

    def test_get(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtailadmin/workflows/create_task.html')

    def test_get_with_non_task_model(self):
        response = self.get(url_kwargs={'app_label': 'wagtailcore', 'model_name': 'Site'})
        self.assertEqual(response.status_code, 404)

    def test_get_with_base_task_model(self):
        response = self.get(url_kwargs={'app_label': 'wagtailcore', 'model_name': 'Task'})
        self.assertEqual(response.status_code, 404)

    def test_post(self):
        response = self.post({'name': 'test_task', 'active': 'on'})

        # Should redirect back to index
        self.assertRedirects(response, reverse('wagtailadmin_workflows:task_index'))

        # Check that the task was created
        tasks = Task.objects.filter(name="test_task", active=True)
        self.assertEqual(tasks.count(), 1)

    def test_permissions(self):
        self.login(user=self.editor)
        response = self.get()
        self.assertEqual(response.status_code, 302)
        full_context = {key: value for context in response.context for key, value in context.items()}
        self.assertEqual(full_context['message'], 'Sorry, you do not have permission to access this area.')

        self.login(user=self.moderator)
        response = self.get()
        self.assertEqual(response.status_code, 200)


class TestSelectTaskTypeView(TestCase, WagtailTestUtils):

    def setUp(self):
        delete_existing_workflows()
        self.login()

    def get(self):
        return self.client.get(reverse('wagtailadmin_workflows:select_task_type'))

    def test_get(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtailadmin/workflows/select_task_type.html')

        # Check that the list of available task types includes SimpleTask and GroupApprovalTask
        self.assertContains(response, SimpleTask.get_verbose_name())
        self.assertContains(response, GroupApprovalTask.get_verbose_name())
        self.assertContains(response, GroupApprovalTask.get_description())


class TestEditTaskView(TestCase, WagtailTestUtils):

    def setUp(self):
        delete_existing_workflows()
        self.login()
        self.task = GroupApprovalTask.objects.create(name="test_task")

        self.editor = self.create_user(
            username='editor',
            email='editor@email.com',
            password='password',
        )
        editors = Group.objects.get(name='Editors')
        editors.user_set.add(self.editor)

        self.moderator = self.create_user(
            username='moderator',
            email='moderator@email.com',
            password='password',
        )
        moderators = Group.objects.get(name='Moderators')
        moderators.user_set.add(self.moderator)
        moderators.permissions.add(Permission.objects.get(codename="change_task"))

    def get(self, params={}):
        return self.client.get(reverse('wagtailadmin_workflows:edit_task', args=[self.task.id]), params)

    def post(self, post_data={}):
        return self.client.post(reverse('wagtailadmin_workflows:edit_task', args=[self.task.id]), post_data)

    def test_get(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtailadmin/workflows/edit_task.html')

    def test_post(self):
        self.assertEqual(self.task.groups.count(), 0)
        editors = Group.objects.get(name='Editors')

        response = self.post({'name': 'test_task_modified', 'active': 'on', 'groups': [str(editors.id)]})

        # Should redirect back to index
        self.assertRedirects(response, reverse('wagtailadmin_workflows:task_index'))

        # Check that the task was updated
        task = GroupApprovalTask.objects.get(id=self.task.id)

        # The task name cannot be changed
        self.assertEqual(task.name, "test_task")

        # This request should've added a group to the task
        self.assertEqual(task.groups.count(), 1)
        self.assertTrue(task.groups.filter(id=editors.id).exists())

    def test_permissions(self):
        self.login(user=self.editor)
        response = self.get()
        self.assertEqual(response.status_code, 302)
        full_context = {key: value for context in response.context for key, value in context.items()}
        self.assertEqual(full_context['message'], 'Sorry, you do not have permission to access this area.')

        self.login(user=self.moderator)
        response = self.get()
        self.assertEqual(response.status_code, 200)

    def test_admin_url_finder(self):
        editor_url_finder = AdminURLFinder(self.editor)
        self.assertEqual(editor_url_finder.get_edit_url(self.task), None)
        moderator_url_finder = AdminURLFinder(self.moderator)
        expected_url = '/admin/workflows/tasks/edit/%d/' % self.task.pk
        self.assertEqual(moderator_url_finder.get_edit_url(self.task), expected_url)


class TestSubmitToWorkflow(TestCase, WagtailTestUtils):
    def setUp(self):
        delete_existing_workflows()
        self.submitter = self.create_user(
            username='submitter',
            email='submitter@email.com',
            password='password',
        )
        editors = Group.objects.get(name='Editors')
        editors.user_set.add(self.submitter)
        self.moderator = self.create_user(
            username='moderator',
            email='moderator@email.com',
            password='password',
        )
        moderators = Group.objects.get(name='Moderators')
        moderators.user_set.add(self.moderator)

        self.superuser = self.create_superuser(
            username='superuser',
            email='superuser@email.com',
            password='password',
        )

        self.login(user=self.submitter)

        # Create a page
        root_page = Page.objects.get(id=2)
        self.page = SimplePage(
            title="Hello world!",
            slug='hello-world',
            content="hello",
            live=False,
            has_unpublished_changes=True,
        )
        root_page.add_child(instance=self.page)
        self.page.save_revision()

        self.workflow, self.task_1, self.task_2 = self.create_workflow_and_tasks()

        WorkflowPage.objects.create(workflow=self.workflow, page=self.page)

    def create_workflow_and_tasks(self):
        workflow = Workflow.objects.create(name='test_workflow')
        task_1 = GroupApprovalTask.objects.create(name='test_task_1')
        task_2 = GroupApprovalTask.objects.create(name='test_task_2')
        task_1.groups.set(Group.objects.filter(name='Moderators'))
        task_2.groups.set(Group.objects.filter(name='Moderators'))
        WorkflowTask.objects.create(workflow=workflow, task=task_1, sort_order=1)
        WorkflowTask.objects.create(workflow=workflow, task=task_2, sort_order=2)
        return workflow, task_1, task_2

    def submit(self):
        post_data = {
            'title': str(self.page.title),
            'slug': str(self.page.slug),
            'content': str(self.page.content),
            'action-submit': "True",
        }
        return self.client.post(reverse('wagtailadmin_pages:edit', args=(self.page.id,)), post_data)

    def test_submit_for_approval_creates_states(self):
        """Test that WorkflowState and TaskState objects are correctly created when a Page is submitted for approval"""

        self.submit()

        workflow_state = self.page.current_workflow_state

        self.assertEqual(type(workflow_state), WorkflowState)
        self.assertEqual(workflow_state.workflow, self.workflow)
        self.assertEqual(workflow_state.status, workflow_state.STATUS_IN_PROGRESS)
        self.assertEqual(workflow_state.requested_by, self.submitter)

        task_state = workflow_state.current_task_state

        self.assertEqual(type(task_state), TaskState)
        self.assertEqual(task_state.task.specific, self.task_1)
        self.assertEqual(task_state.status, task_state.STATUS_IN_PROGRESS)

    def test_submit_for_approval_changes_status_in_header_meta(self):
        edit_url = reverse('wagtailadmin_pages:edit', args=(self.page.id, ))

        response = self.client.get(edit_url)
        self.assertContains(response, 'Draft', count=1)

        # submit for approval
        self.submit()

        response = self.client.get(edit_url)
        workflow_status_url = reverse('wagtailadmin_pages:workflow_status', args=(self.page.id, ))
        self.assertContains(response, workflow_status_url)
        self.assertRegex(response.content.decode('utf-8'), r'Awaiting[\s|\n]+{}'.format(self.page.current_workflow_task.name))
        self.assertNotContains(response, 'Draft')

    def test_submit_sends_mail(self):
        self.submit()
        # 3 emails sent:
        # - to moderator - submitted for approval in moderation stage test_task_1
        # - to superuser - submitted for approval in moderation stage test_task_1
        # - to superuser - submitted to workflow test_workflow
        self.assertEqual(len(mail.outbox), 3)

        # the 'submitted to workflow' email should include the submitter's name
        workflow_message = None
        for msg in mail.outbox:
            if msg.subject == 'The page "Hello world! (simple page)" has been submitted to workflow "test_workflow"':
                workflow_message = msg
                break

        self.assertTrue(workflow_message)
        self.assertIn(
            'The page "Hello world! (simple page)" has been submitted for moderation to workflow "test_workflow" by submitter',
            workflow_message.body
        )

    @mock.patch.object(EmailMultiAlternatives, 'send', side_effect=IOError('Server down'))
    def test_email_send_error(self, mock_fn):
        logging.disable(logging.CRITICAL)

        response = self.submit()
        logging.disable(logging.NOTSET)

        # An email that fails to send should return a message rather than crash the page
        self.assertEqual(response.status_code, 302)
        response = self.client.get(reverse('wagtailadmin_home'))

    def test_resume_rejected_workflow(self):
        # test that an existing workflow can be resumed by submitting when rejected
        self.workflow.start(self.page, user=self.submitter)
        workflow_state = self.page.current_workflow_state
        workflow_state.current_task_state.approve(user=self.superuser)
        workflow_state.refresh_from_db()
        workflow_state.current_task_state.reject(user=self.superuser)
        workflow_state.refresh_from_db()
        self.assertEqual(workflow_state.current_task_state.task.specific, self.task_2)
        self.assertEqual(workflow_state.status, WorkflowState.STATUS_NEEDS_CHANGES)

        self.submit()
        workflow_state.refresh_from_db()

        # check that the same workflow state's status is now in progress
        self.assertEqual(workflow_state.status, WorkflowState.STATUS_IN_PROGRESS)

        # check that the workflow remains on the rejecting task, rather than resetting
        self.assertEqual(workflow_state.current_task_state.task.specific, self.task_2)

    def test_restart_rejected_workflow(self):
        # test that an existing workflow can be restarted when rejected
        self.workflow.start(self.page, user=self.submitter)
        workflow_state = self.page.current_workflow_state
        workflow_state.current_task_state.approve(user=self.superuser)
        workflow_state.refresh_from_db()
        workflow_state.current_task_state.reject(user=self.superuser)
        workflow_state.refresh_from_db()
        self.assertEqual(workflow_state.current_task_state.task.specific, self.task_2)
        self.assertEqual(workflow_state.status, WorkflowState.STATUS_NEEDS_CHANGES)

        post_data = {
            'title': str(self.page.title),
            'slug': str(self.page.slug),
            'content': str(self.page.content),
            'action-restart-workflow': "True",
        }
        self.client.post(reverse('wagtailadmin_pages:edit', args=(self.page.id,)), post_data)
        workflow_state.refresh_from_db()

        # check that the same workflow state's status is now cancelled
        self.assertEqual(workflow_state.status, WorkflowState.STATUS_CANCELLED)

        # check that the new workflow has started on the first task
        new_workflow_state = self.page.current_workflow_state
        self.assertEqual(new_workflow_state.status, WorkflowState.STATUS_IN_PROGRESS)
        self.assertEqual(new_workflow_state.current_task_state.task.specific, self.task_1)

    def test_cancel_workflow(self):
        # test that an existing workflow can be cancelled after submission by the submitter
        self.workflow.start(self.page, user=self.submitter)
        workflow_state = self.page.current_workflow_state
        self.assertEqual(workflow_state.current_task_state.task.specific, self.task_1)
        self.assertEqual(workflow_state.status, WorkflowState.STATUS_IN_PROGRESS)
        post_data = {
            'title': str(self.page.title),
            'slug': str(self.page.slug),
            'content': str(self.page.content),
            'action-cancel-workflow': "True",
        }
        self.client.post(reverse('wagtailadmin_pages:edit', args=(self.page.id,)), post_data)
        workflow_state.refresh_from_db()

        # check that the workflow state's status is now cancelled
        self.assertEqual(workflow_state.status, WorkflowState.STATUS_CANCELLED)
        self.assertEqual(workflow_state.current_task_state.status, TaskState.STATUS_CANCELLED)

    def test_email_headers(self):
        # Submit
        self.submit()

        msg_headers = set(mail.outbox[0].message().items())
        headers = {('Auto-Submitted', 'auto-generated')}
        self.assertTrue(headers.issubset(msg_headers), msg='Message is missing the Auto-Submitted header.',)


@freeze_time("2020-03-31 12:00:00")
class TestApproveRejectWorkflow(TestCase, WagtailTestUtils):
    def setUp(self):
        delete_existing_workflows()
        self.submitter = self.create_user(
            username='submitter',
            first_name='Sebastian',
            last_name='Mitter',
            email='submitter@email.com',
            password='password',
        )
        editors = Group.objects.get(name='Editors')
        editors.user_set.add(self.submitter)
        self.moderator = self.create_user(
            username='moderator',
            email='moderator@email.com',
            password='password',
        )
        moderators = Group.objects.get(name='Moderators')
        moderators.user_set.add(self.moderator)

        self.superuser = self.create_superuser(
            username='superuser',
            email='superuser@email.com',
            password='password',
        )

        self.login(user=self.submitter)

        # Create a page
        root_page = Page.objects.get(id=2)
        self.page = SimplePage(
            title="Hello world!",
            slug='hello-world',
            content="hello",
            live=False,
            has_unpublished_changes=True,
        )
        root_page.add_child(instance=self.page)

        self.workflow, self.task_1 = self.create_workflow_and_tasks()

        WorkflowPage.objects.create(workflow=self.workflow, page=self.page)

        self.submit()

        self.login(user=self.moderator)

    def create_workflow_and_tasks(self):
        workflow = Workflow.objects.create(name='test_workflow')
        task_1 = GroupApprovalTask.objects.create(name='test_task_1')
        task_1.groups.set(Group.objects.filter(name='Moderators'))
        WorkflowTask.objects.create(workflow=workflow, task=task_1, sort_order=1)
        return workflow, task_1

    def submit(self):
        post_data = {
            'title': str(self.page.title),
            'slug': str(self.page.slug),
            'content': str(self.page.content),
            'action-submit': "True",
        }
        return self.client.post(reverse('wagtailadmin_pages:edit', args=(self.page.id,)), post_data)

    @override_settings(WAGTAIL_FINISH_WORKFLOW_ACTION='')
    def test_approve_task_and_workflow(self):
        """
        This posts to the approve task view and checks that the page was approved and published
        """
        # Unset WAGTAIL_FINISH_WORKFLOW_ACTION - default action should be to publish
        del settings.WAGTAIL_FINISH_WORKFLOW_ACTION
        # Connect a mock signal handler to page_published signal
        mock_handler = mock.MagicMock()
        page_published.connect(mock_handler)

        # Post
        self.client.post(reverse('wagtailadmin_pages:workflow_action', args=(self.page.id, 'approve', self.page.current_workflow_task_state.id)), {'comment': 'my comment'})

        # Check that the workflow was approved

        workflow_state = WorkflowState.objects.get(page=self.page, requested_by=self.submitter)

        self.assertEqual(workflow_state.status, workflow_state.STATUS_APPROVED)

        # Check that the task was approved

        task_state = workflow_state.current_task_state

        self.assertEqual(task_state.status, task_state.STATUS_APPROVED)

        # Check that the comment was added to the task state correctly

        self.assertEqual(task_state.comment, 'my comment')

        page = Page.objects.get(id=self.page.id)
        # Page must be live
        self.assertTrue(page.live, "Approving moderation failed to set live=True")
        # Page should now have no unpublished changes
        self.assertFalse(
            page.has_unpublished_changes,
            "Approving moderation failed to set has_unpublished_changes=False"
        )

        # Check that the page_published signal was fired
        self.assertEqual(mock_handler.call_count, 1)
        mock_call = mock_handler.mock_calls[0][2]

        self.assertEqual(mock_call['sender'], self.page.specific_class)
        self.assertEqual(mock_call['instance'], self.page)
        self.assertIsInstance(mock_call['instance'], self.page.specific_class)

    def test_workflow_action_get(self):
        """
        This tests that a GET request to the workflow action view (for the approve action) returns a modal with a form for extra data entry:
        adding a comment
        """
        response = self.client.get(reverse('wagtailadmin_pages:workflow_action', args=(self.page.id, 'approve', self.page.current_workflow_task_state.id)))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtailadmin/pages/workflow_action_modal.html')
        html = json.loads(response.content)['html']
        self.assertTagInHTML('<form action="' + reverse('wagtailadmin_pages:workflow_action', args=(self.page.id, 'approve', self.page.current_workflow_task_state.id)) + '" method="POST" novalidate>', html)
        self.assertIn('Comment', html)

    def test_workflow_action_view_bad_page_id(self):
        """
        This tests that the workflow action view handles invalid page ids correctly
        """
        # Post
        response = self.client.post(reverse('wagtailadmin_pages:workflow_action', args=(127777777777, 'approve', self.page.current_workflow_task_state.id)))

        # Check that the user received a 404 response
        self.assertEqual(response.status_code, 404)

    def test_workflow_action_view_not_in_group(self):
        """
        This tests that the workflow action view for a GroupApprovalTask won't allow approval from a user not in the
        specified group/a superuser
        """
        # Remove privileges from user
        self.login(user=self.submitter)

        # Post
        response = self.client.post(reverse('wagtailadmin_pages:workflow_action', args=(self.page.id, 'approve', self.page.current_workflow_task_state.id)))
        # Check that the user received a permission denied response
        self.assertRedirects(response, '/admin/')

    def test_edit_view_workflow_cancellation_not_in_group(self):
        """
        This tests that the page edit view for a GroupApprovalTask, locked to a user not in the
        specified group/a superuser, still allows the submitter to cancel workflows
        """
        self.login(user=self.submitter)

        # Post
        response = self.client.post(reverse('wagtailadmin_pages:edit', args=(self.page.id, )), {'action-cancel-workflow': 'True'})

        # Check that the user received a 200 response
        self.assertEqual(response.status_code, 200)

        # Check that the workflow state was marked as cancelled
        workflow_state = WorkflowState.objects.get(page=self.page, requested_by=self.submitter)
        self.assertEqual(workflow_state.status, WorkflowState.STATUS_CANCELLED)

    def test_reject_task_and_workflow(self):
        """
        This posts to the reject task view and checks that the page was rejected and not published
        """
        # Post
        self.client.post(reverse('wagtailadmin_pages:workflow_action', args=(self.page.id, 'reject', self.page.current_workflow_task_state.id)))

        # Check that the workflow was marked as needing changes

        workflow_state = WorkflowState.objects.get(page=self.page, requested_by=self.submitter)

        self.assertEqual(workflow_state.status, workflow_state.STATUS_NEEDS_CHANGES)

        # Check that the task was rejected

        task_state = workflow_state.current_task_state

        self.assertEqual(task_state.status, task_state.STATUS_REJECTED)

        page = Page.objects.get(id=self.page.id)
        # Page must not be live
        self.assertFalse(page.live)

    def test_workflow_action_view_rejection_not_in_group(self):
        """
        This tests that the workflow action view for a GroupApprovalTask won't allow rejection from a user not in the
        specified group/a superuser
        """
        # Remove privileges from user
        self.login(user=self.submitter)

        # Post
        response = self.client.post(reverse('wagtailadmin_pages:workflow_action', args=(self.page.id, 'reject', self.page.current_workflow_task_state.id)))

        # Check that the user received a permission denied response
        self.assertRedirects(response, '/admin/')

    def test_collect_workflow_action_data_get(self):
        """
        This tests that a GET request to the collect_workflow_action_data view (for the approve action) returns a modal with a form for extra data entry:
        adding a comment
        """
        response = self.client.get(reverse('wagtailadmin_pages:collect_workflow_action_data', args=(self.page.id, 'approve', self.page.current_workflow_task_state.id)))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtailadmin/pages/workflow_action_modal.html')
        self.assertTemplateUsed(response, 'wagtailadmin/shared/non_field_errors.html')
        html = json.loads(response.content)['html']
        self.assertTagInHTML('<form action="' + reverse('wagtailadmin_pages:collect_workflow_action_data', args=(self.page.id, 'approve', self.page.current_workflow_task_state.id)) + '" method="POST" novalidate>', html)
        self.assertIn('Comment', html)

    def test_collect_workflow_action_data_post(self):
        """
        This tests that a POST request to the collect_workflow_action_data view (for the approve action) returns a modal response with the validated data
        """
        response = self.client.post(
            reverse('wagtailadmin_pages:collect_workflow_action_data', args=(self.page.id, 'approve', self.page.current_workflow_task_state.id)),
            {'comment': "This is my comment"}
        )
        self.assertEqual(response.status_code, 200)
        response_json = json.loads(response.content)
        self.assertEqual(response_json['step'], 'success')
        self.assertEqual(response_json['cleaned_data'], {'comment': "This is my comment"})

    def test_workflow_action_via_edit_view(self):
        """
        Posting to the 'edit' view with 'action-workflow-action' set should perform the given workflow action in addition to updating page content
        """
        # Post
        self.client.post(reverse('wagtailadmin_pages:edit', args=(self.page.id,)), {
            'title': "This title was edited while approving",
            'slug': str(self.page.slug),
            'content': str(self.page.content),
            'action-workflow-action': "True",
            'workflow-action-name': 'approve',
            'workflow-action-extra-data': '{"comment": "my comment"}'
        })

        # Check that the workflow was approved

        workflow_state = WorkflowState.objects.get(page=self.page, requested_by=self.submitter)

        self.assertEqual(workflow_state.status, workflow_state.STATUS_APPROVED)

        # Check that the task was approved

        task_state = workflow_state.current_task_state

        self.assertEqual(task_state.status, task_state.STATUS_APPROVED)

        # Check that the comment was added to the task state correctly

        self.assertEqual(task_state.comment, 'my comment')

        # Check that page edits made at the same time as the action have been saved
        page = Page.objects.get(id=self.page.id)
        self.assertEqual(page.get_latest_revision_as_page().title, "This title was edited while approving")

    def test_workflow_report(self):
        response = self.client.get(reverse('wagtailadmin_reports:workflow'))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Hello world!")
        self.assertContains(response, "test_workflow")
        self.assertContains(response, "Sebastian Mitter")
        self.assertContains(response, "March 31, 2020")

        response = self.client.get(reverse('wagtailadmin_reports:workflow_tasks'))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Hello world!")

    def test_workflow_report_filtered(self):
        # the moderator can review the task, so the workflow state should show up even when reports are filtered by reviewable
        response = self.client.get(reverse('wagtailadmin_reports:workflow'), {'reviewable': 'true'})
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Hello world!")
        self.assertContains(response, "test_workflow")
        self.assertContains(response, "Sebastian Mitter")
        self.assertContains(response, "March 31, 2020")

        response = self.client.get(reverse('wagtailadmin_reports:workflow_tasks'), {'reviewable': 'true'})
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Hello world!")

        # the submitter cannot review the task, so the workflow state shouldn't show up when reports are filtered by reviewable
        self.login(self.submitter)
        response = self.client.get(reverse('wagtailadmin_reports:workflow'), {'reviewable': 'true'})
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(response, "Hello world!")
        self.assertNotContains(response, "Sebastian Mitter")
        self.assertNotContains(response, "March 31, 2020")

        response = self.client.get(reverse('wagtailadmin_reports:workflow_tasks'), {'reviewable': 'true'})
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(response, "Hello world!")


class TestNotificationPreferences(TestCase, WagtailTestUtils):
    def setUp(self):
        delete_existing_workflows()
        self.submitter = self.create_user(
            username='submitter',
            email='submitter@email.com',
            password='password',
        )
        editors = Group.objects.get(name='Editors')
        editors.user_set.add(self.submitter)
        self.moderator = self.create_user(
            username='moderator',
            email='moderator@email.com',
            password='password',
        )
        self.moderator2 = self.create_user(
            username='moderator2',
            email='moderator2@email.com',
            password='password',
        )
        moderators = Group.objects.get(name='Moderators')
        moderators.user_set.add(self.moderator)
        moderators.user_set.add(self.moderator2)

        self.superuser = self.create_superuser(
            username='superuser',
            email='superuser@email.com',
            password='password',
        )

        self.superuser_profile = UserProfile.get_for_user(self.superuser)
        self.moderator2_profile = UserProfile.get_for_user(self.moderator2)
        self.submitter_profile = UserProfile.get_for_user(self.submitter)

        # Create a page
        root_page = Page.objects.get(id=2)
        self.page = SimplePage(
            title="Hello world!",
            slug='hello-world',
            content="hello",
            live=False,
            has_unpublished_changes=True,
        )
        root_page.add_child(instance=self.page)

        self.workflow, self.task_1 = self.create_workflow_and_tasks()

        WorkflowPage.objects.create(workflow=self.workflow, page=self.page)

    def create_workflow_and_tasks(self):
        workflow = Workflow.objects.create(name='test_workflow')
        task_1 = GroupApprovalTask.objects.create(name='test_task_1')
        task_1.groups.set(Group.objects.filter(name='Moderators'))
        WorkflowTask.objects.create(workflow=workflow, task=task_1, sort_order=1)
        return workflow, task_1

    def submit(self):
        post_data = {
            'title': str(self.page.title),
            'slug': str(self.page.slug),
            'content': str(self.page.content),
            'action-submit': "True",
        }
        return self.client.post(reverse('wagtailadmin_pages:edit', args=(self.page.id,)), post_data)

    def approve(self):
        return self.client.post(reverse('wagtailadmin_pages:workflow_action', args=(self.page.id, 'approve', self.page.current_workflow_task_state.id)))

    def reject(self):
        return self.client.post(reverse('wagtailadmin_pages:workflow_action', args=(self.page.id, 'reject', self.page.current_workflow_task_state.id)))

    def test_vanilla_profile(self):
        # Check that the vanilla profile has rejected notifications on
        self.assertEqual(self.submitter_profile.rejected_notifications, True)

        # Check that the vanilla profile has approved notifications on
        self.assertEqual(self.submitter_profile.approved_notifications, True)

    @override_settings(WAGTAILADMIN_NOTIFICATION_INCLUDE_SUPERUSERS=True)
    def test_submitted_email_notifications_sent(self):
        """Test that 'submitted' notifications for WorkflowState and TaskState are both sent correctly"""
        self.login(self.submitter)
        self.submit()

        self.assertEqual(len(mail.outbox), 4)

        task_submission_emails = [email for email in mail.outbox if "task" in email.subject]
        task_submission_emailed_addresses = [address for email in task_submission_emails for address in email.to]
        workflow_submission_emails = [email for email in mail.outbox if "workflow" in email.subject]
        workflow_submission_emailed_addresses = [address for email in workflow_submission_emails for address in email.to]

        self.assertEqual(len(task_submission_emails), 3)
        # the moderator is in the Group assigned to the GroupApproval task, so should get an email
        self.assertIn(self.moderator.email, task_submission_emailed_addresses)
        self.assertIn(self.moderator2.email, task_submission_emailed_addresses)
        # with `WAGTAILADMIN_NOTIFICATION_INCLUDE_SUPERUSERS`, the superuser should get a task email
        self.assertIn(self.superuser.email, task_submission_emailed_addresses)
        # the submitter triggered this workflow update, so should not get an email
        self.assertNotIn(self.submitter.email, task_submission_emailed_addresses)

        self.assertEqual(len(workflow_submission_emails), 1)
        # the moderator should not get a workflow email
        self.assertNotIn(self.moderator.email, workflow_submission_emailed_addresses)
        self.assertNotIn(self.moderator2.email, workflow_submission_emailed_addresses)
        # with `WAGTAILADMIN_NOTIFICATION_INCLUDE_SUPERUSERS`, the superuser should get a workflow email
        self.assertIn(self.superuser.email, workflow_submission_emailed_addresses)
        # as the submitter was the triggering user, the submitter should not get an email notification
        self.assertNotIn(self.submitter.email, workflow_submission_emailed_addresses)

    @override_settings(WAGTAILADMIN_NOTIFICATION_INCLUDE_SUPERUSERS=False)
    def test_submitted_email_notifications_superuser_settings(self):
        """Test that 'submitted' notifications for WorkflowState and TaskState are not sent to superusers if
        `WAGTAILADMIN_NOTIFICATION_INCLUDE_SUPERUSERS=False`"""
        self.login(self.submitter)
        self.submit()

        task_submission_emails = [email for email in mail.outbox if "task" in email.subject]
        task_submission_emailed_addresses = [address for email in task_submission_emails for address in email.to]
        workflow_submission_emails = [email for email in mail.outbox if "workflow" in email.subject]
        workflow_submission_emailed_addresses = [address for email in workflow_submission_emails for address in email.to]

        # with `WAGTAILADMIN_NOTIFICATION_INCLUDE_SUPERUSERS` off, the superuser should not get a task email
        self.assertNotIn(self.superuser.email, task_submission_emailed_addresses)

        # with `WAGTAILADMIN_NOTIFICATION_INCLUDE_SUPERUSERS` off, the superuser should not get a workflow email
        self.assertNotIn(self.superuser.email, workflow_submission_emailed_addresses)

    @override_settings(WAGTAILADMIN_NOTIFICATION_INCLUDE_SUPERUSERS=True)
    def test_submit_notification_active_users_only(self):
        # moderator2 is inactive
        self.moderator2.is_active = False
        self.moderator2.save()

        # superuser is inactive
        self.superuser.is_active = False
        self.superuser.save()

        # Submit
        self.login(self.submitter)
        self.submit()

        workflow_submission_emails = [email for email in mail.outbox if "workflow" in email.subject]
        workflow_submission_emailed_addresses = [address for email in workflow_submission_emails for address in
                                                 email.to]
        task_submission_emails = [email for email in mail.outbox if "task" in email.subject]
        task_submission_emailed_addresses = [address for email in task_submission_emails for address in email.to]

        # Check that moderator2 didn't receive a task submitted email
        self.assertNotIn(self.moderator2.email, task_submission_emailed_addresses)

        # Check that the superuser didn't receive a workflow or task email
        self.assertNotIn(self.superuser.email, task_submission_emailed_addresses)
        self.assertNotIn(self.superuser.email, workflow_submission_emailed_addresses)

    @override_settings(WAGTAILADMIN_NOTIFICATION_INCLUDE_SUPERUSERS=True)
    def test_submit_notification_preferences_respected(self):
        # moderator2 doesn't want emails
        self.moderator2_profile.submitted_notifications = False
        self.moderator2_profile.save()

        # superuser doesn't want emails
        self.superuser_profile.submitted_notifications = False
        self.superuser_profile.save()

        # Submit
        self.login(self.submitter)
        self.submit()

        # Check that only one moderator got a task submitted email
        workflow_submission_emails = [email for email in mail.outbox if "workflow" in email.subject]
        workflow_submission_emailed_addresses = [address for email in workflow_submission_emails for address in email.to]
        task_submission_emails = [email for email in mail.outbox if "task" in email.subject]
        task_submission_emailed_addresses = [address for email in task_submission_emails for address in email.to]
        self.assertNotIn(self.moderator2.email, task_submission_emailed_addresses)

        # Check that the superuser didn't receive a workflow or task email
        self.assertNotIn(self.superuser.email, task_submission_emailed_addresses)
        self.assertNotIn(self.superuser.email, workflow_submission_emailed_addresses)

    def test_approved_notifications(self):
        self.login(self.submitter)
        self.submit()
        # Approve
        self.login(self.moderator)
        self.approve()

        # Submitter must receive a workflow approved email
        workflow_approved_emails = [email for email in mail.outbox if ("workflow" in email.subject and "approved" in email.subject)]
        self.assertEqual(len(workflow_approved_emails), 1)
        self.assertIn(self.submitter.email, workflow_approved_emails[0].to)

    def test_approved_notifications_preferences_respected(self):
        # Submitter doesn't want 'approved' emails
        self.submitter_profile.approved_notifications = False
        self.submitter_profile.save()

        self.login(self.submitter)
        self.submit()
        # Approve
        self.login(self.moderator)
        self.approve()

        # Submitter must not receive a workflow approved email, so there should be no emails in workflow_approved_emails
        workflow_approved_emails = [email for email in mail.outbox if ("workflow" in email.subject and "approved" in email.subject)]
        self.assertEqual(len(workflow_approved_emails), 0)

    def test_rejected_notifications(self):
        self.login(self.submitter)
        self.submit()
        # Reject
        self.login(self.moderator)
        self.reject()

        # Submitter must receive a workflow rejected email
        workflow_rejected_emails = [email for email in mail.outbox if ("workflow" in email.subject and "rejected" in email.subject)]
        self.assertEqual(len(workflow_rejected_emails), 1)
        self.assertIn(self.submitter.email, workflow_rejected_emails[0].to)

    def test_rejected_notification_preferences_respected(self):
        # Submitter doesn't want 'rejected' emails
        self.submitter_profile.rejected_notifications = False
        self.submitter_profile.save()

        self.login(self.submitter)
        self.submit()
        # Reject
        self.login(self.moderator)
        self.reject()

        # Submitter must not receive a workflow rejected email
        workflow_rejected_emails = [email for email in mail.outbox if ("workflow" in email.subject and "rejected" in email.subject)]
        self.assertEqual(len(workflow_rejected_emails), 0)


class TestDisableViews(TestCase, WagtailTestUtils):
    def setUp(self):
        delete_existing_workflows()
        self.submitter = self.create_user(
            username='submitter',
            email='submitter@email.com',
            password='password',
        )
        editors = Group.objects.get(name='Editors')
        editors.user_set.add(self.submitter)
        self.moderator = self.create_user(
            username='moderator',
            email='moderator@email.com',
            password='password',
        )
        self.moderator2 = self.create_user(
            username='moderator2',
            email='moderator2@email.com',
            password='password',
        )
        moderators = Group.objects.get(name='Moderators')
        moderators.user_set.add(self.moderator)
        moderators.user_set.add(self.moderator2)

        self.superuser = self.create_superuser(
            username='superuser',
            email='superuser@email.com',
            password='password',
        )

        # Create a page
        root_page = Page.objects.get(id=2)
        self.page = SimplePage(
            title="Hello world!",
            slug='hello-world',
            content="hello",
            live=False,
            has_unpublished_changes=True,
        )
        root_page.add_child(instance=self.page)

        self.workflow, self.task_1, self.task_2 = self.create_workflow_and_tasks()

        WorkflowPage.objects.create(workflow=self.workflow, page=self.page)

    def create_workflow_and_tasks(self):
        workflow = Workflow.objects.create(name='test_workflow')
        task_1 = GroupApprovalTask.objects.create(name='test_task_1')
        task_1.groups.set(Group.objects.filter(name='Moderators'))
        task_2 = GroupApprovalTask.objects.create(name='test_task_2')
        task_2.groups.set(Group.objects.filter(name='Moderators'))
        WorkflowTask.objects.create(workflow=workflow, task=task_1, sort_order=1)
        WorkflowTask.objects.create(workflow=workflow, task=task_2, sort_order=2)
        return workflow, task_1, task_2

    def submit(self):
        post_data = {
            'title': str(self.page.title),
            'slug': str(self.page.slug),
            'content': str(self.page.content),
            'action-submit': "True",
        }
        return self.client.post(reverse('wagtailadmin_pages:edit', args=(self.page.id,)), post_data)

    def approve(self):
        return self.client.post(reverse('wagtailadmin_pages:workflow_action', args=(self.page.id, 'approve', self.page.current_workflow_task_state.id)))

    def test_disable_workflow(self):
        """Test that deactivating a workflow sets it to inactive and cancels in progress states"""
        self.login(self.submitter)
        self.submit()
        self.login(self.superuser)
        self.approve()

        response = self.client.post(reverse('wagtailadmin_workflows:disable', args=(self.workflow.pk,)))
        self.assertEqual(response.status_code, 302)
        self.workflow.refresh_from_db()
        self.assertEqual(self.workflow.active, False)
        states = WorkflowState.objects.filter(page=self.page, workflow=self.workflow)
        self.assertEqual(states.filter(status=WorkflowState.STATUS_IN_PROGRESS).count(), 0)
        self.assertEqual(states.filter(status=WorkflowState.STATUS_CANCELLED).count(), 1)

        self.assertEqual(TaskState.objects.filter(workflow_state__workflow=self.workflow, status=TaskState.STATUS_IN_PROGRESS).count(), 0)

    def test_disable_task_view(self):
        """Test that a view is shown before disabling a task that shows a warning"""
        self.login(self.submitter)
        self.submit()
        self.login(self.superuser)

        response = self.client.get(reverse('wagtailadmin_workflows:disable_task', args=(self.task_1.pk,)))

        self.assertTemplateUsed(response, "wagtailadmin/workflows/confirm_disable_task.html")
        self.assertEqual(response.context['warning_message'], "This task is in progress on 1 page. Disabling this task will cause it to be skipped in the moderation workflow and not be listed for selection when editing a workflow.")

        # create a new, unused, task and check the warning message is accurate
        unused_task = GroupApprovalTask.objects.create(name='unused_task_3')
        unused_task.groups.set(Group.objects.filter(name='Moderators'))

        response = self.client.get(reverse('wagtailadmin_workflows:disable_task', args=(unused_task.pk,)))

        self.assertEqual(response.context['warning_message'], "This task is in progress on 0 pages. Disabling this task will cause it to be skipped in the moderation workflow and not be listed for selection when editing a workflow.")

        unused_task.delete()  # clean up

    def test_disable_task(self):
        """Test that deactivating a task sets it to inactive and cancels in progress states"""
        self.login(self.submitter)
        self.submit()
        self.login(self.superuser)

        response = self.client.post(reverse('wagtailadmin_workflows:disable_task', args=(self.task_1.pk,)))
        self.assertEqual(response.status_code, 302)
        self.task_1.refresh_from_db()
        self.assertEqual(self.task_1.active, False)
        states = TaskState.objects.filter(workflow_state__page=self.page, task=self.task_1.task_ptr)
        self.assertEqual(states.filter(status=TaskState.STATUS_IN_PROGRESS).count(), 0)
        self.assertEqual(states.filter(status=TaskState.STATUS_CANCELLED).count(), 1)

        # Check that the page's WorkflowState has moved on to the next active task
        self.assertEqual(self.page.current_workflow_state.current_task_state.task.specific, self.task_2)

    def test_enable_workflow(self):
        self.login(self.superuser)
        self.workflow.active = False
        self.workflow.save()

        response = self.client.post(reverse('wagtailadmin_workflows:enable', args=(self.workflow.pk,)))
        self.assertEqual(response.status_code, 302)
        self.workflow.refresh_from_db()
        self.assertEqual(self.workflow.active, True)

    def test_enable_task(self):
        self.login(self.superuser)
        self.task_1.active = False
        self.task_1.save()

        response = self.client.post(reverse('wagtailadmin_workflows:enable_task', args=(self.task_1.pk,)))
        self.assertEqual(response.status_code, 302)
        self.task_1.refresh_from_db()
        self.assertEqual(self.task_1.active, True)


class TestTaskChooserView(TestCase, WagtailTestUtils):
    def setUp(self):
        self.login()

        self.task_enabled = GroupApprovalTask.objects.create(name='Enabled foo')
        self.task_disabled = GroupApprovalTask.objects.create(name='Disabled foo', active=False)

    def test_get(self):
        response = self.client.get(reverse('wagtailadmin_workflows:task_chooser'))

        self.assertEqual(response.status_code, 200)

        self.assertTemplateUsed(response, "wagtailadmin/workflows/task_chooser/chooser.html")

        # Check that the "select task type" view was shown in the "new" tab
        self.assertTemplateUsed(response, "wagtailadmin/workflows/task_chooser/includes/select_task_type.html")
        self.assertTemplateUsed(response, "wagtailadmin/workflows/task_chooser/includes/results.html")
        self.assertTemplateNotUsed(response, "wagtailadmin/workflows/task_chooser/includes/create_form.html")
        self.assertFalse(response.context['searchform'].is_searching())
        # check that only active (non-disabled) tasks are listed
        self.assertEqual([task.name for task in response.context['tasks'].object_list], ['Enabled foo', 'Moderators approval'])

    def test_search(self):
        response = self.client.get(reverse('wagtailadmin_workflows:task_chooser') + '?q=foo')

        self.assertEqual(response.status_code, 200)

        self.assertTemplateUsed(response, "wagtailadmin/workflows/task_chooser/includes/results.html")
        self.assertTemplateNotUsed(response, "wagtailadmin/workflows/task_chooser/chooser.html")
        self.assertTrue(response.context['searchform'].is_searching())
        self.assertEqual(response.context['query_string'], 'foo')
        # check that only active (non-disabled) tasks are listed
        self.assertEqual([task.name for task in response.context['tasks'].object_list], ['Enabled foo'])

    def test_pagination(self):
        response = self.client.get(reverse('wagtailadmin_workflows:task_chooser') + '?p=2')

        self.assertEqual(response.status_code, 200)

        # When pagination is used, only the results template should be rendered
        self.assertTemplateUsed(response, "wagtailadmin/workflows/task_chooser/includes/results.html")
        self.assertTemplateNotUsed(response, "wagtailadmin/workflows/task_chooser/chooser.html")
        self.assertFalse(response.context['searchform'].is_searching())

    def test_get_with_create_model_selected(self):
        response = self.client.get(reverse('wagtailadmin_workflows:task_chooser') + '?create_model=wagtailcore.GroupApprovalTask')

        self.assertEqual(response.status_code, 200)

        self.assertTemplateUsed(response, "wagtailadmin/workflows/task_chooser/chooser.html")

        # Check that the "create" view was shown in the "new" tab
        self.assertTemplateUsed(response, "wagtailadmin/workflows/task_chooser/includes/create_form.html")
        self.assertTemplateNotUsed(response, "wagtailadmin/workflows/task_chooser/includes/select_task_type.html")

    def test_get_with_non_task_create_model_selected(self):
        response = self.client.get(reverse('wagtailadmin_workflows:task_chooser') + '?create_model=wagtailcore.Page')

        self.assertEqual(response.status_code, 404)

    def test_get_with_base_task_create_model_selected(self):
        # Task is technically a subclass of itself so we need an extra test for it
        response = self.client.get(reverse('wagtailadmin_workflows:task_chooser') + '?create_model=wagtailcore.Task')

        self.assertEqual(response.status_code, 404)

    @mock.patch('wagtail.admin.views.workflows.get_task_types')
    def test_get_with_single_task_model(self, get_task_types):
        # When a single task type exists there's no need to specify create_model
        get_task_types.return_value = [GroupApprovalTask]

        response = self.client.get(reverse('wagtailadmin_workflows:task_chooser'))

        self.assertEqual(response.status_code, 200)

        self.assertTemplateUsed(response, "wagtailadmin/workflows/task_chooser/chooser.html")

        # Check that the "create" view was shown in the "new" tab
        self.assertTemplateUsed(response, "wagtailadmin/workflows/task_chooser/includes/create_form.html")
        self.assertTemplateNotUsed(response, "wagtailadmin/workflows/task_chooser/includes/select_task_type.html")

    # POST requests are for creating new tasks

    def get_post_data(self):
        return {
            'create-task-name': 'Editor approval task',
            'create-task-groups': [str(Group.objects.get(name="Editors").id)],
        }

    def test_post_with_create_model_selected(self):
        response = self.client.post(reverse('wagtailadmin_workflows:task_chooser') + '?create_model=wagtailcore.GroupApprovalTask', self.get_post_data())

        self.assertEqual(response.status_code, 201)

        # Check that the task was created
        task = Task.objects.get(name="Editor approval task", active=True)

        # Check the response JSON
        self.assertEqual(response.json(), {
            "step": "task_chosen",
            "result": {
                "id": task.id,
                "name": "Editor approval task",
                "edit_url": reverse('wagtailadmin_workflows:edit_task', args=[task.id])
            }
        })

    @mock.patch('wagtail.admin.views.workflows.get_task_types')
    def test_post_with_single_task_model(self, get_task_types):
        # When a single task type exists there's no need to specify create_model
        get_task_types.return_value = [GroupApprovalTask]

        response = self.client.post(reverse('wagtailadmin_workflows:task_chooser'), self.get_post_data())

        self.assertEqual(response.status_code, 201)

        # Check that the task was created
        task = Task.objects.get(name="Editor approval task", active=True)

        # Check the response JSON
        self.assertEqual(response.json(), {
            "step": "task_chosen",
            "result": {
                "id": task.id,
                "name": "Editor approval task",
                "edit_url": reverse('wagtailadmin_workflows:edit_task', args=[task.id])
            }
        })

    def test_post_without_create_model_selected(self):
        response = self.client.post(reverse('wagtailadmin_workflows:task_chooser'), self.get_post_data())

        self.assertEqual(response.status_code, 400)

        # Check that the task wasn't created
        self.assertFalse(Task.objects.filter(name="Editor approval task", active=True).exists())

    def test_post_with_non_task_create_model_selected(self):
        response = self.client.post(reverse('wagtailadmin_workflows:task_chooser') + '?create_model=wagtailcore.Page', self.get_post_data())

        self.assertEqual(response.status_code, 404)

        # Check that the task wasn't created
        self.assertFalse(Task.objects.filter(name="Editor approval task", active=True).exists())

    def test_post_with_base_task_create_model_selected(self):
        # Task is technically a subclass of itself so we need an extra test for it
        response = self.client.post(reverse('wagtailadmin_workflows:task_chooser') + '?create_model=wagtailcore.Task', self.get_post_data())

        self.assertEqual(response.status_code, 404)

        # Check that the task wasn't created
        self.assertFalse(Task.objects.filter(name="Editor approval task", active=True).exists())


class TestTaskChooserChosenView(TestCase, WagtailTestUtils):
    def setUp(self):
        delete_existing_workflows()
        self.login()
        self.task = SimpleTask.objects.create(name="test_task")

    def test_get(self):
        response = self.client.get(reverse('wagtailadmin_workflows:task_chosen', args=[self.task.id]))

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['Content-Type'], 'application/json')
        self.assertEqual(response.json(), {
            'result': {
                'edit_url': reverse('wagtailadmin_workflows:edit_task', args=[self.task.id]),
                'id': self.task.id,
                'name': 'test_task'
            },
            'step': 'task_chosen'
        })


class TestWorkflowUsageView(TestCase, WagtailTestUtils):
    def setUp(self):
        self.login()
        self.workflow = Workflow.objects.get()

        self.root_page = Page.objects.get(depth=1)
        self.home_page = Page.objects.get(depth=2)

        self.child_page_with_another_workflow = self.home_page.add_child(instance=SimplePage(title="Another page", content="I'm another page"))
        self.another_workflow = Workflow.objects.create(name="Another workflow")
        self.another_workflow.workflow_pages.create(page=self.child_page_with_another_workflow)

    def test_get(self):
        response = self.client.get(reverse('wagtailadmin_workflows:usage', args=[self.workflow.id]))

        self.assertEqual(response.status_code, 200)

        object_set = set(page.id for page in response.context['used_by'].object_list)
        self.assertIn(self.root_page.id, object_set)
        self.assertIn(self.home_page.id, object_set)
        self.assertNotIn(self.child_page_with_another_workflow.id, object_set)


@freeze_time("2020-06-01 12:00:00")
class TestWorkflowStatus(TestCase, WagtailTestUtils):
    def setUp(self):
        delete_existing_workflows()
        self.submitter = self.create_user(
            username='submitter',
            email='submitter@email.com',
            password='password',
        )
        editors = Group.objects.get(name='Editors')
        editors.user_set.add(self.submitter)
        self.moderator = self.create_user(
            username='moderator',
            email='moderator@email.com',
            password='password',
        )
        moderators = Group.objects.get(name='Moderators')
        moderators.user_set.add(self.moderator)

        self.superuser = self.create_superuser(
            username='superuser',
            email='superuser@email.com',
            password='password',
        )

        self.login(self.superuser)

        # Create a page
        root_page = Page.objects.get(id=2)
        self.page = SimplePage(
            title="Hello world!",
            slug='hello-world',
            content="hello",
            live=False,
            has_unpublished_changes=True,
        )
        root_page.add_child(instance=self.page)

        self.workflow, self.task_1, self.task_2 = self.create_workflow_and_tasks()

        WorkflowPage.objects.create(workflow=self.workflow, page=self.page)

        self.edit_url = reverse('wagtailadmin_pages:edit', args=(self.page.id, ))

    def create_workflow_and_tasks(self):
        workflow = Workflow.objects.create(name='test_workflow')
        task_1 = GroupApprovalTask.objects.create(name='test_task_1')
        task_1.groups.set(Group.objects.filter(name='Moderators'))
        WorkflowTask.objects.create(workflow=workflow, task=task_1, sort_order=1)

        task_2 = GroupApprovalTask.objects.create(name='test_task_2')
        task_2.groups.set(Group.objects.filter(name='Editors'))
        WorkflowTask.objects.create(workflow=workflow, task=task_2, sort_order=2)
        return workflow, task_1, task_2

    def submit(self, action='action-submit'):
        post_data = {
            'title': str(self.page.title),
            'slug': str(self.page.slug),
            'content': str(self.page.content),
            action: "True",
        }
        return self.client.post(self.edit_url, post_data)

    def workflow_action(self, action):
        post_data = {
            'action': action,
            'comment': 'good work' if action == 'approve' else 'needs some changes',
            'next': self.edit_url
        }
        return self.client.post(
            reverse('wagtailadmin_pages:workflow_action', args=(self.page.id, action, self.page.current_workflow_task_state.id)),
            post_data,
            follow=True
        )

    def test_workflow_status_modal(self):
        workflow_status_url = reverse('wagtailadmin_pages:workflow_status', args=(self.page.id, ))

        # The page workflow status view should return permission denied when the page is but a draft
        response = self.client.get(workflow_status_url)
        self.assertRedirects(response, '/admin/')

        # Submit for moderation
        self.submit()

        response = self.client.get(workflow_status_url)
        self.assertEqual(response.status_code, 200)
        html = response.json().get('html')
        self.assertIn(self.task_1.name, html)
        self.assertIn('{}: In progress'.format(self.task_1.name), html)
        self.assertIn(self.task_2.name, html)
        self.assertIn('{}: Not started'.format(self.task_2.name), html)
        self.assertIn(reverse('wagtailadmin_pages:history', args=(self.page.id, )), html)

        self.assertTemplateUsed(response, 'wagtailadmin/workflows/workflow_status.html')

    def test_status_through_workflow_cycle(self):
        self.login(self.superuser)
        response = self.client.get(self.edit_url)
        self.assertContains(response, 'Draft', 1)

        self.page.save_revision()
        response = self.client.get(self.edit_url)
        self.assertContains(response, 'Draft saved', 1)

        self.submit()
        response = self.client.get(self.edit_url)
        self.assertRegex(response.content.decode('utf-8'), r'Awaiting[\s|\n]+{}'.format(self.task_1.name))

        response = self.workflow_action('approve')
        self.assertRegex(response.content.decode('utf-8'), r'Awaiting[\s|\n]+{}'.format(self.task_2.name))

        response = self.workflow_action('reject')
        self.assertContains(response, 'Changes requested')

        # resubmit
        self.submit()
        response = self.client.get(self.edit_url)
        self.assertRegex(response.content.decode('utf-8'), r'Awaiting[\s|\n]+{}'.format(self.task_2.name))

        response = self.workflow_action('approve')
        self.assertContains(response, 'Published')

    def test_status_after_cancel(self):
        # start workflow, then cancel
        self.submit()
        self.submit('action-cancel-workflow')
        response = self.client.get(self.edit_url)
        self.assertContains(response, 'Draft saved')

    def test_status_after_restart(self):
        self.submit()
        response = self.workflow_action('approve')
        self.assertRegex(response.content.decode('utf-8'), r'Awaiting[\s|\n]+{}'.format(self.task_2.name))
        self.workflow_action('reject')
        self.submit('action-restart-workflow')
        response = self.client.get(self.edit_url)
        self.assertRegex(response.content.decode('utf-8'), r'Awaiting[\s|\n]+{}'.format(self.task_1.name))

    def test_workflow_status_modal_task_comments(self):
        workflow_status_url = reverse('wagtailadmin_pages:workflow_status', args=(self.page.id,))

        self.submit()
        self.workflow_action('reject')

        response = self.client.get(workflow_status_url)
        self.assertIn('needs some changes', response.json().get('html'))

        self.submit()
        self.workflow_action('approve')
        response = self.client.get(workflow_status_url)
        self.assertIn('good work', response.json().get('html'))

    def test_workflow_edit_locked_message(self):
        self.submit()
        self.login(self.submitter)
        response = self.client.get(self.edit_url)

        needle = "This page is awaiting <b>\'test_task_1\'</b> in the <b>\'test_workflow\'</b> workflow. Only reviewers for this task can edit the page."
        self.assertTagInHTML(needle, str(response.content), count=1)

        self.login(self.moderator)
        response = self.client.get(self.edit_url)
        self.assertNotInHTML(needle, str(response.content))



from django.contrib.auth.models import Permission
from django.urls import include, path, reverse
from django.utils.translation import gettext_lazy as _

from wagtail import hooks
from wagtail.admin.admin_url_finder import ModelAdminURLFinder, register_admin_url_finder
from wagtail.admin.menu import MenuItem
from wagtail.contrib.redirects import urls
from wagtail.contrib.redirects.permissions import permission_policy

from .models import Redirect


@hooks.register('register_admin_urls')
def register_admin_urls():
    return [
        path('redirects/', include(urls, namespace='wagtailredirects')),
    ]


class RedirectsMenuItem(MenuItem):
    def is_shown(self, request):
        return permission_policy.user_has_any_permission(
            request.user, ['add', 'change', 'delete']
        )


@hooks.register('register_settings_menu_item')
def register_redirects_menu_item():
    return RedirectsMenuItem(
        _('Redirects'), reverse('wagtailredirects:index'), icon_name='redirect', order=800
    )


@hooks.register('register_permissions')
def register_permissions():
    return Permission.objects.filter(content_type__app_label='wagtailredirects',
                                     codename__in=['add_redirect', 'change_redirect', 'delete_redirect'])


class RedirectAdminURLFinder(ModelAdminURLFinder):
    edit_url_name = 'wagtailredirects:edit'
    permission_policy = permission_policy


register_admin_url_finder(Redirect, RedirectAdminURLFinder)



from django import forms
from django.template.response import TemplateResponse
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

from wagtail.admin import widgets
from wagtail.admin.views.pages.bulk_actions.page_bulk_action import PageBulkAction
from wagtail.models import Page


class BulkMovePageChooser(widgets.AdminPageChooser):
    def __init__(self, target_models=None, can_choose_root=False, user_perms=None, **kwargs):
        self.pages_to_move = kwargs.pop('pages_to_move', [])
        super().__init__(target_models=target_models, can_choose_root=can_choose_root, user_perms=user_perms, **kwargs)

    @widgets.AdminPageChooser.client_options.getter
    def client_options(self):
        return {
            'model_names': self.model_names,
            'can_choose_root': self.can_choose_root,
            'user_perms': self.user_perms,
            'target_pages': self.pages_to_move,
            'match_subclass': False
        }


class MoveForm(forms.Form):

    def __init__(self, *args, **kwargs):
        destination = kwargs.pop('destination')
        target_parent_models = kwargs.pop('target_parent_models')
        pages_to_move = kwargs.pop('pages_to_move')
        super().__init__(*args, **kwargs)
        self.fields['chooser'] = forms.ModelChoiceField(
            initial=destination,
            queryset=Page.objects.all(),
            widget=BulkMovePageChooser(can_choose_root=True, user_perms='move_to', target_models=target_parent_models, pages_to_move=pages_to_move),
            label=_("Select a new parent page"),
        )


class MoveBulkAction(PageBulkAction):
    display_name = _("Move")
    action_type = "move"
    aria_label = _("Move selected pages")
    template_name = "wagtailadmin/pages/bulk_actions/confirm_bulk_move.html"
    action_priority = 10
    form_class = MoveForm
    destination = None

    def __init__(self, request, model):
        super().__init__(request, model)
        self.target_parent_models = set()
        self.pages_to_move = []

    def get_form_kwargs(self):
        ctx = super().get_form_kwargs()
        ctx['destination'] = self.destination or Page.get_first_root_node()
        ctx['target_parent_models'] = self.target_parent_models
        ctx['pages_to_move'] = self.pages_to_move
        return ctx

    def check_perm(self, page):
        return page.permissions_for_user(self.request.user).can_move()

    def get_success_message(self, num_parent_objects, num_child_objects):
        success_message = ngettext(
            "%(num_pages)d page has been moved",
            "%(num_pages)d pages have been moved",
            num_parent_objects
        ) % {
            'num_pages': num_parent_objects
        }
        return success_message

    def object_context(self, obj):
        context = super().object_context(obj)
        context['child_pages'] = context['item'].get_descendants().count()
        return context

    def get_actionable_objects(self):
        objects, objects_without_access = super().get_actionable_objects()
        request = self.request

        if objects:
            self.target_parent_models = set(objects[0].specific_class.allowed_parent_page_models())
            for obj in objects:
                self.target_parent_models.intersection_update(set(obj.specific_class.allowed_parent_page_models()))

        self.pages_to_move = [page.id for page in objects]

        if self.cleaned_form is None:
            if len(self.target_parent_models) == 0:
                return [], {
                    **objects_without_access,
                    'pages_without_common_parent_page': [
                        {'item': page, 'can_edit': page.permissions_for_user(self.request.user).can_edit()}
                        for page in objects
                    ],
                }
            return objects, objects_without_access

        destination = self.cleaned_form.cleaned_data['chooser']
        pages = []
        pages_without_destination_access = []
        pages_with_duplicate_slugs = []

        for page in objects:
            if not page.permissions_for_user(request.user).can_move_to(destination):
                pages_without_destination_access.append(page)
            elif not Page._slug_is_available(page.slug, destination, page=page):
                pages_with_duplicate_slugs.append(page)
            else:
                pages.append(page)

        return pages, {
            **objects_without_access,
            'pages_without_destination_access': [
                {'item': page, 'can_edit': page.permissions_for_user(self.request.user).can_edit()}
                for page in pages_without_destination_access
            ],
            "pages_with_duplicate_slugs": [
                {'item': page, 'can_edit': page.permissions_for_user(self.request.user).can_edit()}
                for page in pages_with_duplicate_slugs
            ],
        }

    def prepare_action(self, pages, pages_without_access):
        request = self.request
        destination = self.cleaned_form.cleaned_data['chooser']
        if pages_without_access['pages_without_destination_access'] or pages_without_access['pages_with_duplicate_slugs']:
            # this will be picked up by the form
            self.destination = destination
            return TemplateResponse(request, self.template_name, {
                'destination': destination,
                **self.get_context_data()
            })

    def get_execution_context(self):
        return {
            **super().get_execution_context(),
            'destination': self.cleaned_form.cleaned_data['chooser'],
        }

    @classmethod
    def execute_action(cls, objects, destination=None, user=None, **kwargs):
        num_parent_objects = 0
        if destination is None:
            return
        for page in objects:
            page.move(destination, pos='last-child', user=user)
            num_parent_objects += 1
        return num_parent_objects, 0



from django.urls import path

from wagtail.contrib.forms.views import (
    DeleteSubmissionsView, FormPagesListView, get_submissions_list_view)


app_name = 'wagtailforms'
urlpatterns = [
    path('', FormPagesListView.as_view(), name='index'),
    path('submissions/<int:page_id>/', get_submissions_list_view, name='list_submissions'),
    path('submissions/<int:page_id>/delete/', DeleteSubmissionsView.as_view(), name='delete_submissions')
]



import json

from django import forms
from django.contrib.admin.utils import quote
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from wagtail.admin.staticfiles import versioned_static
from wagtail.admin.widgets import AdminChooser
from wagtail.admin.widgets.button import ListingButton
from wagtail.utils.telepath import register
from wagtail.widget_adapters import WidgetAdapter


class AdminSnippetChooser(AdminChooser):

    def __init__(self, model, **kwargs):
        self.target_model = model
        name = self.target_model._meta.verbose_name
        self.choose_one_text = _('Choose %s') % name
        self.choose_another_text = _('Choose another %s') % name
        self.link_to_chosen_text = _('Edit this %s') % name

        super().__init__(**kwargs)

    def get_value_data(self, value):
        if value is None:
            return None
        elif isinstance(value, self.target_model):
            instance = value
        else:  # assume instance ID
            instance = self.target_model.objects.get(pk=value)

        app_label = self.target_model._meta.app_label
        model_name = self.target_model._meta.model_name
        quoted_id = quote(instance.pk)
        edit_url = reverse('wagtailsnippets:edit', args=[app_label, model_name, quoted_id])

        return {
            'id': instance.pk,
            'string': str(instance),
            'edit_url': edit_url,
        }

    def render_html(self, name, value_data, attrs):
        value_data = value_data or {}

        original_field_html = super().render_html(name, value_data.get('id'), attrs)
        chooser_url = reverse(
            'wagtailsnippets:choose', args=[self.target_model._meta.app_label, self.target_model._meta.model_name]
        )

        return render_to_string("wagtailsnippets/widgets/snippet_chooser.html", {
            'widget': self,
            'original_field_html': original_field_html,
            'attrs': attrs,
            'value': bool(value_data),  # only used by chooser.html to identify blank values
            'display_title': value_data.get('string', ''),
            'edit_url': value_data.get('edit_url', ''),
            'chooser_url': chooser_url,
        })

    def render_js_init(self, id_, name, value_data):
        return "createSnippetChooser({id});".format(id=json.dumps(id_))

    @cached_property
    def media(self):
        return forms.Media(js=[
            versioned_static('wagtailsnippets/js/snippet-chooser-modal.js'),
            versioned_static('wagtailsnippets/js/snippet-chooser.js'),
        ])


class SnippetChooserAdapter(WidgetAdapter):
    js_constructor = 'wagtail.contrib.snippets.widgets.SnippetChooser'

    def js_args(self, widget):
        return [
            widget.render_html('__NAME__', None, attrs={'id': '__ID__'}),
            widget.id_for_label('__ID__')
        ]

    @cached_property
    def media(self):
        return forms.Media(js=[
            versioned_static('wagtailsnippets/js/snippet-chooser-telepath.js'),
        ])


register(SnippetChooserAdapter(), AdminSnippetChooser)


class SnippetListingButton(ListingButton):
    pass



# -*- coding: utf-8 -*-
from django.db import migrations


def reassign_admin_access_permission(apps, schema_editor):
    ContentType = apps.get_model('contenttypes.ContentType')
    Permission = apps.get_model('auth.Permission')
    Group = apps.get_model('auth.Group')

    # Add a content type to hang the 'can access Wagtail admin' permission off
    admin_content_type, created = ContentType.objects.get_or_create(
        app_label='wagtailcore',
        model='admin'
    )

    try:
        permission = Permission.objects.get(codename='access_admin')
        # wagtail.admin has been used, update the existing permission
        permission.content_type = admin_content_type
        permission.save()
    except Permission.DoesNotExist:
        # wagtail.admin not used, create a new permission

        # Create admin permission
        admin_permission, created = Permission.objects.get_or_create(
            content_type=admin_content_type,
            codename='access_admin',
            name='Can access Wagtail admin'
        )

        # Assign it to Editors and Moderators groups
        for group in Group.objects.filter(name__in=['Editors', 'Moderators']):
            group.permissions.add(admin_permission)


class Migration(migrations.Migration):

    dependencies = [
        ('wagtailcore', '0069_admin'),
    ]

    operations = [
        migrations.RunPython(reassign_admin_access_permission),
    ]



from django.test import TestCase
from django.urls import reverse
from django.utils.http import urlencode

from wagtail.test.testapp.models import EventPage
from wagtail.test.utils import WagtailTestUtils


class TestContentTypeUse(TestCase, WagtailTestUtils):
    fixtures = ['test.json']

    def setUp(self):
        self.user = self.login()
        self.christmas_page = EventPage.objects.get(title="Christmas")

    def test_content_type_use(self):
        # Get use of event page
        request_url = reverse('wagtailadmin_pages:type_use', args=('tests', 'eventpage'))
        response = self.client.get(request_url)

        # Check response
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wagtailadmin/pages/content_type_use.html')
        self.assertContains(response, "Christmas")

        # Links to 'delete' etc should include a 'next' URL parameter pointing back here
        delete_url = (
            reverse('wagtailadmin_pages:delete', args=(self.christmas_page.id,))
            + '?' + urlencode({'next': request_url})
        )
        self.assertContains(response, delete_url)



# Generated by Django 3.1.8 on 2021-06-22 14:19

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('wagtailsearchpromotions', '0002_capitalizeverbose'),
    ]

    operations = [
        migrations.CreateModel(
            name='Query',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('query_string', models.CharField(max_length=255, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name='QueryDailyHits',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateField()),
                ('hits', models.IntegerField(default=0)),
                ('query', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='daily_hits', to='wagtailsearchpromotions.query')),
            ],
            options={
                'verbose_name': 'Query Daily Hits',
                'verbose_name_plural': 'Query Daily Hits',
                'unique_together': {('query', 'date')},
            },
        ),
    ]



# Generated by Django 3.1.3 on 2020-11-27 01:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtailcore', '0059_apply_collection_ordering'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='workflowstate',
            name='unique_in_progress_workflow',
        ),
        migrations.AddConstraint(
            model_name='workflowstate',
            constraint=models.UniqueConstraint(condition=models.Q(status__in=('in_progress', 'needs_changes')), fields=('page',), name='unique_in_progress_workflow'),
        ),
    ]



import json
import logging
import re

from draftjs_exporter.defaults import render_children
from draftjs_exporter.dom import DOM
from draftjs_exporter.html import HTML as HTMLExporter

from wagtail.admin.rich_text.converters.html_to_contentstate import (
    BLOCK_KEY_NAME, HtmlToContentStateHandler)
from wagtail.rich_text import features as feature_registry
from wagtail.utils.whitelist import check_url


def link_entity(props):
    """
    <a linktype="page" id="1">internal page link</a>
    """
    id_ = props.get('id')
    link_props = {}

    if id_ is not None:
        link_props['linktype'] = 'page'
        link_props['id'] = id_
    else:
        link_props['href'] = check_url(props.get('url'))

    return DOM.create_element('a', link_props, props['children'])


def br(props):
    if props['block']['type'] == 'code-block':
        return props['children']

    return DOM.create_element('br')


def block_fallback(props):
    type_ = props['block']['type']
    logging.error('Missing config for "%s". Deleting block.' % type_)
    return None


def entity_fallback(props):
    type_ = props['entity']['type']
    logging.warning('Missing config for "%s". Deleting entity' % type_)
    return None


def style_fallback(props):
    type_ = props['inline_style_range']['style']
    logging.warning('Missing config for "%s". Deleting style.' % type_)
    return props['children']


def persist_key_for_block(config):
    # For any block level element config for draft js exporter, return a config that retains the
    # block key in a data attribute
    if isinstance(config, dict):
        # Wrapper elements don't retain a key - we can keep them in the config as-is
        new_config = {key: value for key, value in config.items() if key in {'wrapper', 'wrapper_props'}}
        element = config.get('element')
        element_props = config.get('props', {})
    else:
        # The config is either a simple string element name, or a function
        new_config = {}
        element_props = {}
        element = config

    def element_with_uuid(props):
        added_props = {BLOCK_KEY_NAME: props['block'].get('key')}
        try:
            # See if the element is a function - if so, we can only run it and modify its return value to include the data attribute
            elt = element(props)
            if elt is not None:
                elt.attr.update(added_props)
            return elt
        except TypeError:
            # Otherwise we can do the normal process of creating a DOM element with the right element type
            # and simply adding the data attribute to its props
            added_props.update(element_props)
            return DOM.create_element(element, added_props, props['children'])

    new_config['element'] = element_with_uuid
    return new_config


class ContentstateConverter():
    def __init__(self, features=None):
        self.features = features
        self.html_to_contentstate_handler = HtmlToContentStateHandler(features)

        exporter_config = {
            'block_map': {
                'unstyled': persist_key_for_block('p'),
                'atomic': render_children,
                'fallback': block_fallback,
            },
            'style_map': {
                'FALLBACK': style_fallback,
            },
            'entity_decorators': {
                'FALLBACK': entity_fallback,
            },
            'composite_decorators': [
                {
                    'strategy': re.compile(r'\n'),
                    'component': br,
                },
            ],
            'engine': DOM.STRING,
        }

        for feature in self.features:
            rule = feature_registry.get_converter_rule('contentstate', feature)
            if rule is not None:
                feature_config = rule['to_database_format']
                exporter_config['block_map'].update({block_type: persist_key_for_block(config) for block_type, config in feature_config.get('block_map', {}).items()})
                exporter_config['style_map'].update(feature_config.get('style_map', {}))
                exporter_config['entity_decorators'].update(feature_config.get('entity_decorators', {}))

        self.exporter = HTMLExporter(exporter_config)

    def from_database_format(self, html):
        self.html_to_contentstate_handler.reset()
        self.html_to_contentstate_handler.feed(html)
        self.html_to_contentstate_handler.close()

        return self.html_to_contentstate_handler.contentstate.as_json(indent=4, separators=(',', ': '))

    def to_database_format(self, contentstate_json):
        return self.exporter.render(json.loads(contentstate_json))



import itertools

from django import forms
from django.core.exceptions import ValidationError
from django.forms.utils import ErrorList
from django.utils.functional import cached_property
from django.utils.html import format_html, format_html_join
from django.utils.translation import gettext as _

from wagtail.admin.staticfiles import versioned_static
from wagtail.utils.telepath import Adapter, register

from .base import Block, get_help_icon


__all__ = ['ListBlock', 'ListBlockValidationError']


class ListBlockValidationError(ValidationError):
    def __init__(self, block_errors=None, non_block_errors=None):
        self.non_block_errors = non_block_errors or ErrorList()
        self.block_errors = block_errors or []

        params = {}
        if block_errors:
            params['block_errors'] = block_errors
        if non_block_errors:
            params['non_block_errors'] = non_block_errors
        super().__init__('Validation error in ListBlock', params=params)


class ListBlockValidationErrorAdapter(Adapter):
    js_constructor = 'wagtail.blocks.ListBlockValidationError'

    def js_args(self, error):
        return [
            [elist.as_data() if elist is not None else elist for elist in error.block_errors],
            error.non_block_errors.as_data(),
        ]

    @cached_property
    def media(self):
        return forms.Media(js=[
            versioned_static('wagtailadmin/js/telepath/blocks.js'),
        ])


register(ListBlockValidationErrorAdapter(), ListBlockValidationError)


class ListBlock(Block):

    def __init__(self, child_block, **kwargs):
        super().__init__(**kwargs)

        if isinstance(child_block, type):
            # child_block was passed as a class, so convert it to a block instance
            self.child_block = child_block()
        else:
            self.child_block = child_block

        if not hasattr(self.meta, 'default'):
            # Default to a list consisting of one empty (i.e. default-valued) child item
            self.meta.default = [self.child_block.get_default()]

    def get_default(self):
        # wrap with list() so that each invocation of get_default returns a distinct instance
        return list(self.meta.default)

    def value_from_datadict(self, data, files, prefix):
        count = int(data['%s-count' % prefix])
        values_with_indexes = []
        for i in range(0, count):
            if data['%s-%d-deleted' % (prefix, i)]:
                continue
            values_with_indexes.append(
                (
                    int(data['%s-%d-order' % (prefix, i)]),
                    self.child_block.value_from_datadict(data, files, '%s-%d-value' % (prefix, i))
                )
            )

        values_with_indexes.sort()
        return [v for (i, v) in values_with_indexes]

    def value_omitted_from_data(self, data, files, prefix):
        return ('%s-count' % prefix) not in data

    def clean(self, value):
        result = []
        errors = []
        non_block_errors = ErrorList()
        for child_val in value:
            try:
                result.append(self.child_block.clean(child_val))
            except ValidationError as e:
                errors.append(ErrorList([e]))
            else:
                errors.append(None)

        if self.meta.min_num is not None and self.meta.min_num > len(value):
            non_block_errors.append(ValidationError(
                _('The minimum number of items is %d') % self.meta.min_num
            ))

        if self.meta.max_num is not None and self.meta.max_num < len(value):
            non_block_errors.append(ValidationError(
                _('The maximum number of items is %d') % self.meta.max_num
            ))

        if any(errors) or non_block_errors:
            raise ListBlockValidationError(block_errors=errors, non_block_errors=non_block_errors)

        return result

    def to_python(self, value):
        # 'value' is a list of child block values; use bulk_to_python to convert them all in one go
        return self.child_block.bulk_to_python(value)

    def bulk_to_python(self, values):
        # 'values' is a list of lists of child block values; concatenate them into one list so that
        # we can make a single call to child_block.bulk_to_python
        lengths = [len(val) for val in values]
        raw_values = list(itertools.chain.from_iterable(values))
        converted_values = self.child_block.bulk_to_python(raw_values)

        # split converted_values back into sub-lists of the original lengths
        result = []
        offset = 0
        for sublist_len in lengths:
            result.append(converted_values[offset:offset + sublist_len])
            offset += sublist_len

        return result

    def get_prep_value(self, value):
        # recursively call get_prep_value on children and return as a list
        return [
            self.child_block.get_prep_value(item)
            for item in value
        ]

    def get_form_state(self, value):
        return [
            self.child_block.get_form_state(item)
            for item in value
        ]

    def get_api_representation(self, value, context=None):
        # recursively call get_api_representation on children and return as a list
        return [
            self.child_block.get_api_representation(item, context=context)
            for item in value
        ]

    def render_basic(self, value, context=None):
        children = format_html_join(
            '\n', '<li>{0}</li>',
            [
                (self.child_block.render(child_value, context=context),)
                for child_value in value
            ]
        )
        return format_html("<ul>{0}</ul>", children)

    def get_searchable_content(self, value):
        content = []

        for child_value in value:
            content.extend(self.child_block.get_searchable_content(child_value))

        return content

    def check(self, **kwargs):
        errors = super().check(**kwargs)
        errors.extend(self.child_block.check(**kwargs))
        return errors

    class Meta:
        # No icon specified here, because that depends on the purpose that the
        # block is being used for. Feel encouraged to specify an icon in your
        # descendant block type
        icon = "placeholder"
        form_classname = None
        min_num = None
        max_num = None

    MUTABLE_META_ATTRIBUTES = ['min_num', 'max_num']


class ListBlockAdapter(Adapter):
    js_constructor = 'wagtail.blocks.ListBlock'

    def js_args(self, block):
        meta = {
            'label': block.label, 'icon': block.meta.icon, 'classname': block.meta.form_classname,
            'strings': {
                'MOVE_UP': _("Move up"),
                'MOVE_DOWN': _("Move down"),
                'DUPLICATE': _("Duplicate"),
                'DELETE': _("Delete"),
                'ADD': _("Add"),
            },
        }
        help_text = getattr(block.meta, 'help_text', None)
        if help_text:
            meta['helpText'] = help_text
            meta['helpIcon'] = get_help_icon()

        if block.meta.min_num is not None:
            meta['minNum'] = block.meta.min_num

        if block.meta.max_num is not None:
            meta['maxNum'] = block.meta.max_num

        return [
            block.name,
            block.child_block,
            block.child_block.get_form_state(block.child_block.get_default()),
            meta,
        ]

    @cached_property
    def media(self):
        return forms.Media(js=[
            versioned_static('wagtailadmin/js/telepath/blocks.js'),
        ])


register(ListBlockAdapter(), ListBlock)


DECONSTRUCT_ALIASES = {
    ListBlock: 'wagtail.blocks.ListBlock',
}



from datetime import timedelta

import django_filters

from django import forms
from django.core.exceptions import PermissionDenied
from django.core.paginator import Paginator
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _

from wagtail.admin.auth import user_has_any_page_permission, user_passes_test
from wagtail.admin.filters import DateRangePickerWidget, WagtailFilterSet
from wagtail.admin.views.reports import ReportView
from wagtail.logging import registry as log_action_registry
from wagtail.models import Page, PageRevision, UserPagePermissionsProxy, logging, workflows


class PageHistoryReportFilterSet(WagtailFilterSet):
    action = django_filters.ChoiceFilter(choices=log_action_registry.get_choices)
    hide_commenting_actions = django_filters.BooleanFilter(
        label=_('Hide commenting actions'),
        method='filter_hide_commenting_actions',
        widget=forms.CheckboxInput,
    )
    user = django_filters.ModelChoiceFilter(
        field_name='user', queryset=lambda request: logging.PageLogEntry.objects.all().get_users()
    )
    timestamp = django_filters.DateFromToRangeFilter(label=_('Date'), widget=DateRangePickerWidget)

    def filter_hide_commenting_actions(self, queryset, name, value):
        if value:
            queryset = queryset.exclude(
                action__startswith='wagtail.comments'
            )
        return queryset

    class Meta:
        model = logging.PageLogEntry
        fields = ['action', 'user', 'timestamp', 'hide_commenting_actions']


def workflow_history(request, page_id):
    page = get_object_or_404(Page, id=page_id)

    user_perms = UserPagePermissionsProxy(request.user)
    if not user_perms.for_page(page).can_edit():
        raise PermissionDenied

    workflow_states = workflows.WorkflowState.objects.filter(page=page).order_by('-created_at')

    paginator = Paginator(workflow_states, per_page=20)
    workflow_states = paginator.get_page(request.GET.get('p'))

    return TemplateResponse(request, 'wagtailadmin/pages/workflow_history/index.html', {
        'page': page,
        'workflow_states': workflow_states,
    })


def workflow_history_detail(request, page_id, workflow_state_id):
    page = get_object_or_404(Page, id=page_id)

    user_perms = UserPagePermissionsProxy(request.user)
    if not user_perms.for_page(page).can_edit():
        raise PermissionDenied

    workflow_state = get_object_or_404(workflows.WorkflowState, page=page, id=workflow_state_id)

    # Get QuerySet of all revisions that have existed during this workflow state
    # It's possible that the page is edited while the workflow is running, so some
    # tasks may be repeated. All tasks that have been completed no matter what
    # revision needs to be displayed on this page.
    page_revisions = PageRevision.objects.filter(
        page=page,
        id__in=workflows.TaskState.objects.filter(workflow_state=workflow_state).values_list('page_revision_id', flat=True)
    ).order_by('-created_at')

    # Now get QuerySet of tasks completed for each revision
    task_states_by_revision_task = [
        (page_revision, {
            task_state.task: task_state
            for task_state in workflows.TaskState.objects.filter(workflow_state=workflow_state, page_revision=page_revision)
        })
        for page_revision in page_revisions
    ]

    # Make sure task states are always in a consistent order
    # In some cases, they can be completed in a different order to what they are defined
    tasks = workflow_state.workflow.tasks.all()
    task_states_by_revision = [
        (
            page_revision,
            [
                task_states_by_task.get(task, None)
                for task in tasks
            ]
        )
        for page_revision, task_states_by_task in task_states_by_revision_task
    ]

    # Generate timeline
    completed_task_states = workflows.TaskState.objects.filter(
        workflow_state=workflow_state
    ).exclude(
        finished_at__isnull=True
    ).exclude(
        status=workflows.TaskState.STATUS_CANCELLED
    )

    timeline = [
        {
            'time': workflow_state.created_at,
            'action': 'workflow_started',
            'workflow_state': workflow_state,
        }
    ]

    if workflow_state.status not in (workflows.WorkflowState.STATUS_IN_PROGRESS, workflows.WorkflowState.STATUS_NEEDS_CHANGES):
        last_task = completed_task_states.order_by('finished_at').last()
        if last_task:
            timeline.append({
                'time': last_task.finished_at + timedelta(milliseconds=1),
                'action': 'workflow_completed',
                'workflow_state': workflow_state,
            })

    for page_revision in page_revisions:
        timeline.append({
            'time': page_revision.created_at,
            'action': 'page_edited',
            'revision': page_revision,
        })

    for task_state in completed_task_states:
        timeline.append({
            'time': task_state.finished_at,
            'action': 'task_completed',
            'task_state': task_state,
        })

    timeline.sort(key=lambda t: t['time'])
    timeline.reverse()

    return TemplateResponse(request, 'wagtailadmin/pages/workflow_history/detail.html', {
        'page': page,
        'workflow_state': workflow_state,
        'tasks': tasks,
        'task_states_by_revision': task_states_by_revision,
        'timeline': timeline,
    })


class PageHistoryView(ReportView):
    template_name = 'wagtailadmin/pages/history.html'
    title = _('Page history')
    header_icon = 'history'
    paginate_by = 20
    filterset_class = PageHistoryReportFilterSet

    @method_decorator(user_passes_test(user_has_any_page_permission))
    def dispatch(self, request, *args, **kwargs):
        self.page = get_object_or_404(Page, id=kwargs.pop('page_id')).specific

        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, *args, object_list=None, **kwargs):
        context = super().get_context_data(*args, object_list=object_list, **kwargs)
        context['page'] = self.page
        context['subtitle'] = self.page.get_admin_display_title()
        context['page_latest_revision'] = self.page.get_latest_revision()

        return context

    def get_queryset(self):
        return logging.PageLogEntry.objects.filter(page=self.page).select_related(
            'revision',
            'user',
            'user__wagtail_userprofile'
        )



# Generated by Django 3.0.6 on 2020-04-24 13:00

import django.contrib.postgres.indexes
import django.contrib.postgres.search
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('postgres_search', '0002_add_autocomplete'),
    ]

    operations = [
        migrations.AddField(
            model_name='indexentry',
            name='title',
            field=django.contrib.postgres.search.SearchVectorField(default=''),
            preserve_default=False,
        ),
        migrations.AddIndex(
            model_name='indexentry',
            index=django.contrib.postgres.indexes.GinIndex(fields=['title'], name='postgres_se_title_b56f33_gin'),
        ),
        migrations.AddField(
            model_name='indexentry',
            name='title_norm',
            field=models.FloatField(default=1.0),
        ),
    ]



import datetime

from io import BytesIO
from unittest import mock

from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.core import checks
from django.test import TestCase
from django.utils.timezone import make_aware
from openpyxl import load_workbook

from wagtail.admin.admin_url_finder import AdminURLFinder
from wagtail.contrib.images.models import Image
from wagtail.contrib.images.tests.utils import get_test_image_file
from wagtail.contrib.modeladmin.helpers.search import DjangoORMSearchHandler
from wagtail.edit_handlers import FieldPanel, TabbedInterface
from wagtail.models import ModelLogEntry, Page
from wagtail.test.modeladmintest.models import Author, Book, Publisher, RelatedLink, Token
from wagtail.test.modeladmintest.wagtail_hooks import BookModelAdmin
from wagtail.test.utils import WagtailTestUtils


class TestBookIndexView(TestCase, WagtailTestUtils):
    fixtures = ['modeladmintest_test.json']

    def setUp(self):
        self.login()

        img = Image.objects.create(
            title="LOTR cover",
            file=get_test_image_file(),
        )
        book = Book.objects.get(title="The Lord of the Rings")
        book.cover_image = img
        book.save()

    def get(self, **params):
        return self.client.get('/admin/modeladmintest/book/', params)

    def test_simple(self):
        response = self.get()

        self.assertEqual(response.status_code, 200)

        # There are four books in the test data
        self.assertEqual(response.context['result_count'], 4)

        # The result count content is shown in the header
        self.assertContains(response, '<span class="result-count">4 out of 4</span>', html=True)

        # User has add permission
        self.assertEqual(response.context['user_can_create'], True)

    def test_csv_export(self):
        # Export the whole queryset
        response = self.get(export='csv')

        self.assertEqual(response.status_code, 200)
        # Check attachment is present and named correctly using the modeladmin export_filename
        self.assertEqual(response.get('content-disposition'), 'attachment; filename="books-export.csv"')

        # Check response - all books should be in it
        data_lines = response.getvalue().decode().split("\n")
        self.assertEqual(data_lines[0], 'Title,Author,Author Date Of Birth\r')
        self.assertEqual(data_lines[1], 'Charlie and the Chocolate Factory,Roald Dahl,1916-09-13\r')
        self.assertEqual(data_lines[2], 'The Chronicles of Narnia,Roald Dahl,1898-11-29\r')
        self.assertEqual(data_lines[3], 'The Hobbit,J. R. R. Tolkien,1892-01-03\r')
        self.assertEqual(data_lines[4], 'The Lord of the Rings,J. R. R. Tolkien,1892-01-03\r')

    def test_xlsx_export(self):
        # Export the whole queryset
        response = self.get(export='xlsx')

        self.assertEqual(response.status_code, 200)
        # Check attachment is present and named correctly using the modeladmin export_filename
        self.assertEqual(response.get('content-disposition'), 'attachment; filename="books-export.xlsx"')

        # Check response - all books should be in it
        workbook_data = response.getvalue()
        worksheet = load_workbook(filename=BytesIO(workbook_data))['Sheet1']
        cell_array = [[cell.value for cell in row] for row in worksheet.rows]
        self.assertEqual(cell_array[0], ['Title', 'Author', 'Author Date Of Birth'])
        self.assertEqual(cell_array[1], ['Charlie and the Chocolate Factory', 'Roald Dahl', '1916-09-13'])
        self.assertEqual(cell_array[2], ['The Chronicles of Narnia', 'Roald Dahl', '1898-11-29'])
        self.assertEqual(cell_array[3], ['The Hobbit', 'J. R. R. Tolkien', '1892-01-03'])
        self.assertEqual(cell_array[4], ['The Lord of the Rings', 'J. R. R. Tolkien', '1892-01-03'])
        self.assertEqual(len(cell_array), 5)

    def test_tr_attributes(self):
        response = self.get()

        # Charlie & The Chocolate factory should be in the list with the
        # `data-author_yob` and `data-object_pk` attributes added
        self.assertContains(response, 'data-author-yob="1916"')
        self.assertContains(response, 'data-object-pk="3"')

        # There should be two odd rows and two even ones, and 'book' should be
        # added to the `class` attribute for every one.
        self.assertContains(response, 'class="book odd"', count=2)
        self.assertContains(response, 'class="book even"', count=2)

    def test_filter(self):
        # Filter by author 1 (JRR Tolkien)
        response = self.get(author__id__exact=1)

        self.assertEqual(response.status_code, 200)

        # JRR Tolkien has two books in the test data
        self.assertEqual(response.context['result_count'], 2)

        # The result count content is shown in the header
        self.assertContains(response, '<span class="result-count">2 out of 4</span>', html=True)

        for book in response.context['object_list']:
            self.assertEqual(book.author_id, 1)

    def test_filtered_csv_export(self):
        # Filter by author 1 (JRR Tolkien) and export the current selection
        response = self.get(author__id__exact=1, export='csv')

        # Check response - only books by JRR Tolkien should be in it
        self.assertEqual(response.status_code, 200)
        data_lines = response.getvalue().decode().split("\n")
        self.assertEqual(data_lines[0], 'Title,Author,Author Date Of Birth\r')
        self.assertEqual(data_lines[1], 'The Hobbit,J. R. R. Tolkien,1892-01-03\r')
        self.assertEqual(data_lines[2], 'The Lord of the Rings,J. R. R. Tolkien,1892-01-03\r')
        self.assertEqual(data_lines[3], '')

    def test_search_indexed(self):
        response = self.get(q='of')

        self.assertEqual(response.status_code, 200)

        # There are two books where the title contains 'of'
        self.assertEqual(response.context['result_count'], 2)

        # The result count content is shown in the header
        self.assertContains(response, '<span class="result-count">2 out of 4</span>', html=True)

    def test_search_form_present(self):
        # Test the backend search handler allows the search form to render
        response = self.get()

        self.assertContains(response, '<input id="id_q"')

    def test_search_form_absent(self):
        # DjangoORMSearchHandler + no search_fields, search form should be absent
        with mock.patch.object(BookModelAdmin, 'search_handler_class', DjangoORMSearchHandler):
            response = self.get()

            self.assertNotContains(response, '<input id="id_q"')

    def test_ordering(self):
        response = self.get(o='0.1')

        self.assertEqual(response.status_code, 200)

        # There are four books in the test data
        self.assertEqual(response.context['result_count'], 4)

    def test_paging(self):
        # should be corrected to just the first page, as there aren't enough
        # objects to make up more than one page
        response = self.get(p=9)

        self.assertEqual(response.status_code, 200)

        # There are four books in the test data
        self.assertEqual(response.context['result_count'], 4)

        # Should raise a ValueError that gets caught during initialisation
        response = self.get(p='notaninteger')

        self.assertEqual(response.status_code, 200)

        # There are four books in the test data
        self.assertEqual(response.context['result_count'], 4)


class TestAuthorIndexView(TestCase, WagtailTestUtils):
    fixtures = ['modeladmintest_test.json']

    def setUp(self):
        self.login()

    def get(self, **params):
        return self.client.get('/admin/modeladmintest/author/', params)

    def test_search(self):
        response = self.get(q='Roald Dahl')

        self.assertEqual(response.status_code, 200)

        self.assertEqual(response.context['result_count'], 2)

        # The result count content is shown in the header
        self.assertContains(response, '<span class="result-count">2 out of 5</span>', html=True)

    def test_col_extra_class_names(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        test_html = """
            <td class="field-first_book for-author-1 title">The Lord of the Rings</td>
        """
        self.assertContains(response, test_html, html=True)

    def test_col_extra_attributes(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        test_html = """
            <td class="field-last_book title" data-for_author="1">The Hobbit</td>
        """
        self.assertContains(response, test_html, html=True)


class TestCreateView(TestCase, WagtailTestUtils):
    fixtures = ['modeladmintest_test.json']

    def setUp(self):
        self.login()

    def get(self):
        return self.client.get('/admin/modeladmintest/book/create/')

    def post(self, post_data):
        return self.client.post('/admin/modeladmintest/book/create/', post_data)

    def test_simple(self):
        response = self.get()

        self.assertEqual(response.status_code, 200)

    def test_create(self):
        response = self.post({
            'title': "George's Marvellous Medicine",
            'author': 2,
        })
        # Should redirect back to index
        self.assertRedirects(response, '/admin/modeladmintest/book/')

        # Check that the book was created
        self.assertEqual(Book.objects.filter(title="George's Marvellous Medicine").count(), 1)

        response = self.client.get('/admin/modeladmintest/publisher/create/')
        self.assertIn('name', response.content.decode('UTF-8'))
        self.assertNotIn('headquartered_in', response.content.decode('UTF-8'))
        self.assertEqual(
            [ii for ii in response.context['form'].fields],
            ['name']
        )
        self.client.post('/admin/modeladmintest/publisher/create/', {
            'name': 'Sharper Collins'
        })
        publisher = Publisher.objects.get(name='Sharper Collins')
        self.assertEqual(publisher.headquartered_in, None)

    def test_post_invalid(self):
        initial_book_count = Book.objects.count()

        response = self.post({
            'title': '',
            'author': 2,
        })
        final_book_count = Book.objects.count()

        self.assertEqual(response.status_code, 200)
        # Check that the book was not created
        self.assertEqual(initial_book_count, final_book_count)

        # Check that a form error was raised
        self.assertFormError(response, 'form', 'title', "This field is required.")
        self.assertContains(response, """<p class="error-message"><span>This field is required.</span></p>""",
                            count=1, html=True)

    def test_exclude_passed_to_extract_panel_definitions(self):
        path_to_form_fields_exclude_property = 'wagtail.contrib.modeladmin.options.ModelAdmin.form_fields_exclude'
        with mock.patch('wagtail.contrib.modeladmin.options.extract_panel_definitions_from_model_class') as m:
            with mock.patch(path_to_form_fields_exclude_property, new_callable=mock.PropertyMock) as mock_form_fields_exclude:
                mock_form_fields_exclude.return_value = ['123']

                self.get()
                self.assertTrue(mock_form_fields_exclude.called)
                m.assert_called_with(Book, exclude=mock_form_fields_exclude.return_value)

    def test_clean_form_once(self):
        with mock.patch('wagtail.test.modeladmintest.wagtail_hooks.PublisherModelAdminForm.clean') as mock_form_clean:
            response = self.client.post('/admin/modeladmintest/publisher/create/', {'name': ''})
            self.assertEqual(response.status_code, 200)

            mock_form_clean.assert_called_once()

    def test_create_view_with_multifieldpanel(self):
        # https://github.com/wagtail/wagtail/issues/6413
        response = self.client.get('/admin/modeladmintest/relatedlink/create/')
        self.assertEqual(response.status_code, 200)

        response = self.client.post('/admin/modeladmintest/relatedlink/create/', {
            'title': "Homepage",
            'link': Page.objects.filter(depth=2).first().id,
        })
        # Should redirect back to index
        self.assertRedirects(response, '/admin/modeladmintest/relatedlink/')

        # Check that the link was created
        self.assertEqual(RelatedLink.objects.filter(title="Homepage").count(), 1)

    def test_prepopulated_field_data_in_context(self):
        response = self.get()
        self.assertIn('data-prepopulated-fields="[{&quot;id&quot;: &quot;#id_title&quot;, &quot;name&quot;: &quot;title&quot;, &quot;dependency_ids&quot;: [&quot;#id_author&quot;], &quot;dependency_list&quot;: [&quot;author&quot;], &quot;maxLength&quot;: 255, &quot;allowUnicode&quot;: false}]"', response.content.decode('UTF-8'))


class TestInspectView(TestCase, WagtailTestUtils):
    fixtures = ['modeladmintest_test.json']

    def setUp(self):
        self.login()

        img = Image.objects.create(
            title="LOTR cover",
            file=get_test_image_file(),
        )
        book = Book.objects.get(title="The Lord of the Rings")
        book.cover_image = img
        book.save()

    def get_for_author(self, author_id):
        return self.client.get('/admin/modeladmintest/author/inspect/%d/' % author_id)

    def get_for_book(self, book_id):
        return self.client.get('/admin/modeladmintest/book/inspect/%d/' % book_id)

    def test_author_simple(self):
        response = self.get_for_author(1)
        self.assertEqual(response.status_code, 200)

    def test_author_name_present(self):
        """
        The author name should appear twice. Once in the header, and once
        more in the field listing
        """
        response = self.get_for_author(1)
        self.assertContains(response, 'J. R. R. Tolkien', 2)

    def test_author_dob_not_present(self):
        """
        The date of birth shouldn't appear, because the field wasn't included
        in the `inspect_view_fields` list
        """
        response = self.get_for_author(1)
        self.assertNotContains(response, '1892')

    def test_book_simple(self):
        response = self.get_for_book(1)
        self.assertEqual(response.status_code, 200)

    def test_book_title_present(self):
        """
        The book title should appear once only, in the header, as 'title'
        was added to the `inspect_view_fields_ignore` list
        """
        response = self.get_for_book(1)
        self.assertContains(response, 'The Lord of the Rings', 1)

    def test_book_author_present(self):
        """
        The author name should appear, because 'author' is not in
        `inspect_view_fields_ignore` and should be returned by the
        `get_inspect_view_fields` method.
        """
        response = self.get_for_book(1)
        self.assertContains(response, 'J. R. R. Tolkien', 1)

    def test_non_existent(self):
        response = self.get_for_book(100)
        self.assertEqual(response.status_code, 404)


class TestEditView(TestCase, WagtailTestUtils):
    fixtures = ['modeladmintest_test.json']

    def setUp(self):
        self.user = self.login()
        ModelLogEntry.objects.create(
            content_type=ContentType.objects.get_for_model(Book),
            label="The Lord of the Rings",
            action='wagtail.create',
            timestamp=make_aware(datetime.datetime(2021, 9, 30, 10, 1, 0)),
            object_id='1',
        )

    def get(self, book_id):
        return self.client.get('/admin/modeladmintest/book/edit/%d/' % book_id)

    def post(self, book_id, post_data):
        return self.client.post('/admin/modeladmintest/book/edit/%d/' % book_id, post_data)

    def test_simple(self):
        response = self.get(1)

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'The Lord of the Rings')

        # "Last updated" timestamp should be present
        self.assertContains(response, 'data-wagtail-tooltip="Sept. 30, 2021, 10:01 a.m."')
        # History link should be present
        self.assertContains(response, 'href="/admin/modeladmintest/book/history/1/"')

        url_finder = AdminURLFinder(self.user)
        expected_url = '/admin/modeladmintest/book/edit/1/'
        self.assertEqual(url_finder.get_edit_url(Book.objects.get(id=1)), expected_url)

    def test_non_existent(self):
        response = self.get(100)

        self.assertEqual(response.status_code, 404)

    def test_edit(self):
        response = self.post(1, {
            'title': 'The Lady of the Rings',
            'author': 1,
        })

        # Should redirect back to index
        self.assertRedirects(response, '/admin/modeladmintest/book/')

        # Check that the book was updated
        self.assertEqual(Book.objects.get(id=1).title, 'The Lady of the Rings')

    def test_post_invalid(self):
        response = self.post(1, {
            'title': '',
            'author': 1,
        })

        self.assertEqual(response.status_code, 200)

        # Check that the title was not updated
        self.assertEqual(Book.objects.get(id=1).title, 'The Lord of the Rings')

        # Check that a form error was raised
        self.assertFormError(response, 'form', 'title', "This field is required.")
        self.assertContains(response, """<p class="error-message"><span>This field is required.</span></p>""",
                            count=1, html=True)

    def test_exclude_passed_to_extract_panel_definitions(self):
        path_to_form_fields_exclude_property = 'wagtail.contrib.modeladmin.options.ModelAdmin.form_fields_exclude'
        with mock.patch('wagtail.contrib.modeladmin.options.extract_panel_definitions_from_model_class') as m:
            with mock.patch(path_to_form_fields_exclude_property, new_callable=mock.PropertyMock) as mock_form_fields_exclude:
                mock_form_fields_exclude.return_value = ['123']

                self.get(1)
                self.assertTrue(mock_form_fields_exclude.called)
                m.assert_called_with(Book, exclude=mock_form_fields_exclude.return_value)

    def test_clean_form_once(self):
        with mock.patch('wagtail.test.modeladmintest.wagtail_hooks.PublisherModelAdminForm.clean') as mock_form_clean:
            publisher = Publisher.objects.create(name='Sharper Collins')

            response = self.client.post('/admin/modeladmintest/publisher/edit/%d/' % publisher.pk, {'name': ''})
            self.assertEqual(response.status_code, 200)

            mock_form_clean.assert_called_once()

    def test_edit_view_with_multifieldpanel(self):
        # https://github.com/wagtail/wagtail/issues/6413
        link = RelatedLink.objects.create(title='Homepage', link=Page.objects.filter(depth=2).first())
        response = self.client.get('/admin/modeladmintest/relatedlink/edit/%d/' % link.id)
        self.assertEqual(response.status_code, 200)

        response = self.client.post('/admin/modeladmintest/relatedlink/edit/%d/' % link.id, {
            'title': "Homepage edited",
            'link': Page.objects.filter(depth=2).first().id,
        })
        # Should redirect back to index
        self.assertRedirects(response, '/admin/modeladmintest/relatedlink/')

        # Check that the link was updated
        self.assertEqual(RelatedLink.objects.filter(title="Homepage edited").count(), 1)


class TestPageSpecificViews(TestCase, WagtailTestUtils):
    fixtures = ['modeladmintest_test.json']
    expected_status_code = 404

    def setUp(self):
        self.login()

    def test_choose_parent(self):
        response = self.client.get('/admin/modeladmintest/book/choose_parent/')
        self.assertEqual(response.status_code, self.expected_status_code)


class TestConfirmDeleteView(TestCase, WagtailTestUtils):
    fixtures = ['modeladmintest_test.json']

    def setUp(self):
        self.login()

    def get(self, book_id):
        return self.client.get('/admin/modeladmintest/book/delete/%d/' % book_id)

    def post(self, book_id):
        return self.client.post('/admin/modeladmintest/book/delete/%d/' % book_id)

    def test_simple(self):
        response = self.get(1)

        self.assertEqual(response.status_code, 200)

    def test_non_existent(self):
        response = self.get(100)

        self.assertEqual(response.status_code, 404)

    def test_post(self):
        response = self.post(1)

        # User redirected to index
        self.assertRedirects(response, '/admin/modeladmintest/book/')

        # Book deleted
        self.assertFalse(Book.objects.filter(id=1).exists())


class TestDeleteViewWithProtectedRelation(TestCase, WagtailTestUtils):
    fixtures = ['modeladmintest_test.json']

    def setUp(self):
        self.login()

    def get(self, author_id):
        return self.client.get('/admin/modeladmintest/author/delete/%d/' % author_id)

    def post(self, author_id):
        return self.client.post('/admin/modeladmintest/author/delete/%d/' % author_id)

    def test_get_with_dependent_object(self):
        response = self.get(1)

        self.assertEqual(response.status_code, 200)

    def test_get_without_dependent_object(self):
        response = self.get(4)

        self.assertEqual(response.status_code, 200)

    def test_post_with_dependent_object(self):
        response = self.post(1)

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            "'J. R. R. Tolkien' is currently referenced by other objects"
        )
        self.assertContains(
            response,
            "<li><b>Book:</b> The Lord of the Rings</li>"
        )

        # Author not deleted
        self.assertTrue(Author.objects.filter(id=1).exists())

    def test_post_without_dependent_object(self):
        response = self.post(4)

        # User redirected to index
        self.assertRedirects(response, '/admin/modeladmintest/author/')

        # Author deleted
        self.assertFalse(Author.objects.filter(id=4).exists())

    def test_post_with_1to1_dependent_object(self):
        response = self.post(5)

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            "'Harper Lee' is currently referenced by other objects"
        )
        self.assertContains(
            response,
            "<li><b>Solo Book:</b> To Kill a Mockingbird</li>"
        )

        # Author not deleted
        self.assertTrue(Author.objects.filter(id=5).exists())


class TestDeleteViewModelReprPrimary(TestCase, WagtailTestUtils):
    fixtures = ['modeladmintest_test.json']

    def setUp(self):
        self.login()

    def test_delete(self):
        response = self.client.post('/admin/modeladmintest/token/delete/boom/')
        self.assertEqual(response.status_code, 302)


class TestEditorAccess(TestCase, WagtailTestUtils):
    fixtures = ['modeladmintest_test.json']

    def setUp(self):
        # Create a user
        user = self.create_user(username='test2', password='password')
        user.groups.add(Group.objects.get(pk=2))
        self.user = user
        # Login
        self.login(username='test2', password='password')

        return user

    def test_index_permitted(self):
        response = self.client.get('/admin/modeladmintest/book/')
        self.assertRedirects(response, '/admin/')

    def test_inpspect_permitted(self):
        response = self.client.get('/admin/modeladmintest/book/inspect/2/')
        self.assertRedirects(response, '/admin/')

    def test_create_permitted(self):
        response = self.client.get('/admin/modeladmintest/book/create/')
        self.assertRedirects(response, '/admin/')

    def test_edit_permitted(self):
        response = self.client.get('/admin/modeladmintest/book/edit/2/')
        self.assertRedirects(response, '/admin/')

        url_finder = AdminURLFinder(self.user)
        self.assertEqual(url_finder.get_edit_url(Book.objects.get(id=2)), None)

    def test_delete_get_permitted(self):
        response = self.client.get('/admin/modeladmintest/book/delete/2/')
        self.assertRedirects(response, '/admin/')

    def test_delete_post_permitted(self):
        response = self.client.post('/admin/modeladmintest/book/delete/2/')
        self.assertRedirects(response, '/admin/')


class TestHistoryView(TestCase, WagtailTestUtils):
    fixtures = ['modeladmintest_test.json']

    def setUp(self):
        self.login()
        ModelLogEntry.objects.create(
            content_type=ContentType.objects.get_for_model(Book),
            label="The Lord of the Rings",
            action='wagtail.create',
            timestamp=make_aware(datetime.datetime(2021, 9, 30, 10, 1, 0)),
            object_id='1',
        )

    def test_simple(self):
        response = self.client.get('/admin/modeladmintest/book/history/1/')
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<td>Created</td>', html=True)
        self.assertContains(response, '<div class="human-readable-date" title="Sept. 30, 2021, 10:01 a.m.">')


class TestQuoting(TestCase, WagtailTestUtils):
    fixtures = ['modeladmintest_test.json']
    expected_status_code = 200

    def setUp(self):
        self.login()
        self.tok_reg = Token.objects.create(key="RegularName")
        self.tok_irr = Token.objects.create(key="Irregular_Name")

    def test_action_links(self):
        response = self.client.get('/admin/modeladmintest/token/')

        self.assertContains(response, 'href="/admin/modeladmintest/token/edit/RegularName/"')
        self.assertContains(response, 'href="/admin/modeladmintest/token/delete/RegularName/"')
        self.assertContains(response, 'href="/admin/modeladmintest/token/edit/Irregular_5FName/"')
        self.assertContains(response, 'href="/admin/modeladmintest/token/delete/Irregular_5FName/"')

        response = self.client.get('/admin/modeladmintest/token/edit/Irregular_5FName/')
        self.assertEqual(response.status_code, 200)
        response = self.client.get('/admin/modeladmintest/token/delete/Irregular_5FName/')
        self.assertEqual(response.status_code, 200)


class TestHeaderBreadcrumbs(TestCase, WagtailTestUtils):
    """
        Test that the <ul class="breadcrumbs">... is inserted within the
        <header> tag for potential future regression.
        See https://github.com/wagtail/wagtail/issues/3889
    """
    fixtures = ['modeladmintest_test.json']

    def setUp(self):
        self.login()

    def test_choose_inspect_model(self):
        response = self.client.get('/admin/modeladmintest/author/inspect/1/')

        # check correct templates were used
        self.assertTemplateUsed(response, 'modeladmin/includes/breadcrumb.html')
        self.assertTemplateUsed(response, 'wagtailadmin/shared/header.html')

        # check that home breadcrumb link exists
        expected = """
            <li class="home">
                <a href="/admin/">
                    <svg class="icon icon-home home_icon" aria-hidden="true" focusable="false">
                        <use href="#icon-home"></use>
                    </svg>
                    <span class="visuallyhidden">Home</span>
                    <svg class="icon icon-arrow-right arrow_right_icon" aria-hidden="true" focusable="false">
                        <use href="#icon-arrow-right"></use>
                    </svg>
                </a>
            </li>
        """
        self.assertContains(response, expected, html=True)

        # check that the breadcrumbs are before the first header closing tag
        content_str = str(response.content)
        position_of_header_close = content_str.index('</header>')
        position_of_breadcrumbs = content_str.index('<ul class="breadcrumb">')
        self.assertGreater(position_of_header_close, position_of_breadcrumbs)


class TestPanelConfigurationChecks(TestCase, WagtailTestUtils):

    def setUp(self):
        self.warning_id = 'wagtailadmin.W002'

        def get_checks_result():
            # run checks only with the 'panels' tag
            checks_result = checks.run_checks(tags=['panels'])
            return [
                warning for warning in
                checks_result if warning.id == self.warning_id]

        self.get_checks_result = get_checks_result

    def test_model_with_single_tabbed_panel_only(self):
        Publisher.content_panels = [FieldPanel('name'), FieldPanel('headquartered_in')]

        warning = checks.Warning(
            "Publisher.content_panels will have no effect on modeladmin editing",
            hint="""Ensure that Publisher uses `panels` instead of `content_panels`\
or set up an `edit_handler` if you want a tabbed editing interface.
There are no default tabs on non-Page models so there will be no\
 Content tab for the content_panels to render in.""",
            obj=Publisher,
            id='wagtailadmin.W002',
        )

        checks_results = self.get_checks_result()

        self.assertIn(warning, checks_results)

        # clean up for future checks
        delattr(Publisher, 'content_panels')

    def test_model_with_two_tabbed_panels_only(self):
        Publisher.settings_panels = [FieldPanel('name')]
        Publisher.promote_panels = [FieldPanel('headquartered_in')]

        warning_1 = checks.Warning(
            "Publisher.promote_panels will have no effect on modeladmin editing",
            hint="""Ensure that Publisher uses `panels` instead of `promote_panels`\
or set up an `edit_handler` if you want a tabbed editing interface.
There are no default tabs on non-Page models so there will be no\
 Promote tab for the promote_panels to render in.""",
            obj=Publisher,
            id='wagtailadmin.W002',
        )

        warning_2 = checks.Warning(
            "Publisher.settings_panels will have no effect on modeladmin editing",
            hint="""Ensure that Publisher uses `panels` instead of `settings_panels`\
or set up an `edit_handler` if you want a tabbed editing interface.
There are no default tabs on non-Page models so there will be no\
 Settings tab for the settings_panels to render in.""",
            obj=Publisher,
            id='wagtailadmin.W002',
        )

        checks_results = self.get_checks_result()

        self.assertIn(warning_1, checks_results)
        self.assertIn(warning_2, checks_results)

        # clean up for future checks
        delattr(Publisher, 'settings_panels')
        delattr(Publisher, 'promote_panels')

    def test_model_with_single_tabbed_panel_and_edit_handler(self):
        Publisher.content_panels = [FieldPanel('name'), FieldPanel('headquartered_in')]
        Publisher.edit_handler = TabbedInterface(Publisher.content_panels)

        # no errors should occur
        self.assertEqual(self.get_checks_result(), [])

        # clean up for future checks
        delattr(Publisher, 'content_panels')
        delattr(Publisher, 'edit_handler')
